//***************************************************************************
//*  ====================
//*  ZH2LXAnalysis Classes
//*  ====================
//*
//*  (Description)
//*	A user analysis class for JLC analyses.
//*	This reads and analyzes MC e+e- -> ZH data.
//*  (Requires)
//*	library Anlib (in physsim-99a-1 if K.Fujii)
//*	library ZHStudy (also in physsim)
//*  (Provides)
//*	class ZH2LXAnalysis
//*	class ZH2LXAnalysisBuf
//*  (Usage)
//*	...
//*  (Update Record)
//*	23 June 2015	K.Fujii		Original version.
//***************************************************************************

#include "ZH2LXAnalysis.h"
#include "ANLTrack.h"
#include <sstream>
#include <iomanip>

static const Double_t kEcm     = 250.0; // Ecm
static const Double_t kMassH   = 125.0; // H mass
static const Double_t kMassZ   = 91.19;	// Z mass
static const Double_t kSigmaMh =   4.0;	// H mass resolution
static const Double_t kSigmaMz =   2.0;	// Z mass resolution
static const Int_t    kZoneX   =     4;	// No. of X Zones in the Canvas
static const Int_t    kZoneY   =     4;	// No. of Y Zones in the Canvas

static const Double_t kPi      = TMath::ACos(-1.);
static const Double_t k2Pi     = 2.*kPi;

Bool_t gDEBUG = kFALSE;
//Bool_t gDEBUG = kTRUE;
Int_t  gNgoods  = 0;
const Int_t  kMaxCuts = 100;
      stringstream gCutName[kMaxCuts];

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//  ---------------------
//  ZH2LXAnalysisBuf Class
//  ---------------------
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

ClassImp(ZH2LXAnalysisBuf)

// Constructors
ZH2LXAnalysisBuf::ZH2LXAnalysisBuf(const Char_t *name, const Char_t *title,
   ZH2LXAnalysis *module) : JSFEventBuf(name, title, (JSFModule*)module) {}

ZH2LXAnalysisBuf::ZH2LXAnalysisBuf(ZH2LXAnalysis *module, const Char_t *name,
   const Char_t *title) : JSFEventBuf(name, title, (JSFModule*)module) {}

ZH2LXAnalysisBuf::~ZH2LXAnalysisBuf() {}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//  ------------------
//  ZH2LXAnalysis Class
//  ------------------
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

ClassImp(ZH2LXAnalysis)

// Constructor
ZH2LXAnalysis::ZH2LXAnalysis(const Char_t *name, const Char_t *title)
               : JSFModule(name, title)
{
  cHist = 0;
  fEventBuf = new ZH2LXAnalysisBuf(this);
  SetBufferSize(2000);	// buffer size for event data
  cout << "ZH2LXAnalysis is created... fEventBuf is "
       << (void *)fEventBuf << endl;
}

// Destructor
ZH2LXAnalysis::~ZH2LXAnalysis()
{
  cout << "ZH2LXAnalysisBuf will be deleted... fEventBuf is "
       << (void *)fEventBuf << endl;
  delete fEventBuf; fEventBuf = 0;
}


//####### Functions ###########

void ZH2LXAnalysis::CleanUp(TObjArray *objs)
{
  TIter next(objs);
  TObject *obj;
  while ( (obj = next()) ) {
    objs->Remove(obj);
    delete obj;
  }
}

Bool_t ZH2LXAnalysis::Initialize()
{
  TDirectory *last = gDirectory;
  gFile->cd("/");

  hStat       = new TH1F("hStat","Cut Statistics"   ,  20,   0.0,   20.0);
  hNtracks    = new TH1F("hNtracks","No. of Tracks" ,  50,   0.0,  200.0);
  hEvis       = new TH1F("hEvis","Visible Energy"   ,  50,   0.0,  500.0);
  hPt         = new TH1F("hPt","Missing Pt"         ,  50,   0.0,  250.0);
  hNlptracks  = new TH1F("hNlptracks","No.lp trks"  ,  10,   0.0,   10.0);
  hNsols      = new TH1F("hNsols","No. of solutions",  20,   0.0,   20.0);
  hChi2       = new TH1F("hChi2","Chi2"             ,  50,   0.0,   50.0);
  hEvisPl     = new TH2F("hEvisPl","(Evis,Pl)"      ,
                                     60,  0.0, 600.0,  50,-100.0, +100.0);
  hEzEh       = new TH2F("hEzEh","E_Z vs. E_H"      ,
				   200,  0.0, 200.0,  200,   0.0,  200.0);
  hCoszCosh   = new TH2F("hCoszCosh","Cos(theta_Z) vs. Cos(theta_H)",
                                     50, -1.0,  +1.0,  50,  -1.0,   +1.0);
  hAcop       = new TH1F("hAcop","Acoplanarity"     ,  90,   0.0,  180.0);
  hMassh      = new TH1F("hMassh","M_H Distribution", 200,  50.0,  150.0);
  hMassz      = new TH1F("hMassz","M_Z Distribution", 240,  60.0,  120.0);
  hMrec       = new TH1F("hMrec","M_rec Distribution", 800,  90.0, 250.0);
  hCosz       = new TH1F("hCosz","Cos(theta_Z)"     ,  50,  -1.0,   +1.0);
  hPhid       = new TH1F("hPhid","Phi_d"            ,  50,    .0,   k2Pi);

  xNtracks  =      3;	// No. of Tracks
  xEtrack   =   0.05;	// track energy
  xEvis     =  50.00;	// Minimum visible energy
  xPt       = 999.00;	// Pt maximum
  xPl       = 999.00;	// Pl maximum
  xElepton  =  10.00;   // Elepton mimimum
  xCosCone  =   0.98;   // cos(theta_cone)
  xEcone    =   2.00;   // Ecome maximum
  xCoszh    =   1.00;	// |cos(theta_Z)| and |cos(theta_H)| maximum
  xMll      =  18.00;	// |m_ll-m_Z| maximum
  xMrec     =  18.00;	// |m_rec-m_H| maximum
  xAcop     =  30.00;	// Acoplanarity maximum

  last->cd();
  return 0;
}

void ZH2LXAnalysis::DrawHist()
{
  TDirectory *last = gDirectory;
  if (!cHist) {
    cHist = new TCanvas("cHist","Canvas 1", 10, 10, kZoneX*200, kZoneY*200);
    cHist->Divide(kZoneX,kZoneY);
  } 
  else {
    cHist->cd();
  }

  cHist->cd( 1);	hStat->Draw();
  cHist->cd( 2);	hNtracks->Draw();
  cHist->cd( 3);	hEvis->Draw();
  cHist->cd( 4);	hPt->Draw();
  cHist->cd( 6);	hNsols->Draw();
  cHist->cd( 7);	hChi2->Draw();
  cHist->cd( 8);	hEvisPl->Draw();
  cHist->cd( 9);	hEzEh->Draw();
  cHist->cd(10);	hCoszCosh->Draw();
  cHist->cd(11);	hAcop->Draw();
  cHist->cd(12);	hMassz->Draw();
  cHist->cd(13);	hMassh->Draw();
  cHist->cd(14);	hMrec->Draw();
  cHist->cd(15);	hCosz->Draw();
  cHist->cd(16);	hPhid->Draw();

  cHist->Update();

  last->cd();
}


// *****Process()***** //
Bool_t ZH2LXAnalysis::Process(Int_t ev)
{
  // Local copies of ZH2LXAnalysisBuf data members.

  Int_t		fNtracks;	// track multiplicity
  Int_t         fNlptracks;     // Isolated lepton multiplicity
  Double_t	fEvis;		// visible energy
  Double_t	fPt;		// Pt
  Double_t	fPl;		// Pl

  // Remember the previous directory.
  
  TDirectory *last = gDirectory;
  gFile->cd("/");

  // Analysis starts here.

  Float_t selid = -0.5;
  hStat->Fill(++selid);
  if (gNgoods == 0) {
    gCutName[(Int_t)selid] << "No Cuts" << ends;
  }

  // Get event buffer and make combined tracks accessible.

  JSFSIMDST        *sds	= static_cast<JSFSIMDST *>(gJSF->FindModule("JSFSIMDST"));
  JSFSIMDSTBuf	   *evt	= static_cast<JSFSIMDSTBuf*>(sds->EventBuf());
  ZH2LXAnalysisBuf *ua	= static_cast<ZH2LXAnalysisBuf*>(fEventBuf);
  ZH2LXAnalysisBuf &a	= *ua;

  Int_t		ntrks	= evt->GetNLTKCLTracks();	// No. of tracks
  TObjArray	*trks	= evt->GetLTKCLTracks();	// Combined tracks

  // Select good tracks
  ANL4DVector qsum;
  TObjArray tracks(1000);
  fNtracks = 0;
  for ( Int_t i = 0; i < ntrks; i++) {
    JSFLTKCLTrack *t = static_cast<JSFLTKCLTrack*>(trks->UncheckedAt(i));
    if ( t->GetE() > xEtrack ) {
      ANLTrack *qt = new ANLTrack(t);
      tracks.Add(qt);		// track 4-momentum
      qsum += *qt;		// total 4-mometum
      fNtracks++;		// *qt stays.
    }
  }
  if (gDEBUG) cerr << "Ntracks = " << fNtracks << endl;

  // Cut on No. of tracks.
  
  hNtracks->Fill(fNtracks);
  if ( fNtracks < xNtracks ) { CleanUp(&tracks); return kFALSE; }
  hStat->Fill(++selid);
  if (gNgoods == 0) {
    gCutName[(Int_t)selid] << "Ntracks >= " << xNtracks << ends;
  }
  
  ANL4DVector qcm(kEcm,0.,0.,0.);

  fEvis = qsum(0);	// E_vis
  fPt	= qsum.GetPt(); // P_t
  fPl	= qsum(3);	// P_l

  if (gDEBUG) cerr << "Evis = " << fEvis << " Pt = "
	<< fPt << " Pl = " << fPl << endl;

  // Cut on Evis.

  hEvis->Fill(fEvis);
  if ( fEvis < xEvis ) { CleanUp(&tracks); return kFALSE; }
  hStat->Fill(++selid);
  if ( gNgoods == 0 ) {
    gCutName[(Int_t)selid] << "Evis >= " << xEvis << ends;
  }

  // Cut on Pt.
  
  hPt->Fill(fPt);
  if ( fPt > xPt ) { CleanUp(&tracks); return kFALSE; }
  hStat->Fill(++selid);
  if ( gNgoods == 0 ) {
    gCutName[(Int_t)selid] << "Pt <= " << xPt << ends;
  }

  // Cut on Pl.
  
  if ( TMath::Abs(fPl) > xPl ) { CleanUp(&tracks); return kFALSE; }
  hStat->Fill(++selid);
  if ( gNgoods == 0 ) {
    gCutName[(Int_t)selid] << "|Pl| <= " << xPl << ends;
  }

  // Find Isolated Lepton.

  TObjArray lptracks(20);
  fNlptracks = 0;
  TIter nexttrk (&tracks);
  ANLTrack *trkp;
  while ((trkp = static_cast<ANLTrack *>(nexttrk()))) {
    ANLTrack &trk = *trkp;
    if ( ! trk.IsLepton() ) continue;
    Double_t elepton = trk.E();
    if ( elepton < xElepton ) continue;
    Double_t econe = trk.GetConeEnergy(xCosCone, &tracks);
    if ( econe <= xEcone ) {
      lptracks.Add(trkp);
      fNlptracks++;
    }
  }

  // Require at least two Isolated Leptons.

  hNlptracks->Fill(fNlptracks);
  if ( fNlptracks < 2 ) { CleanUp(&tracks); return kFALSE; }
  hStat->Fill(++selid);
  if ( gNgoods == 0 ) {
    gCutName[(Int_t)selid] << "Nlptracks >= 2" << ends;
  }

  // Find ZH candidates in given mass window.
  
  TObjArray solutions(10);
  ANLPairCombiner zcandidates(lptracks,lptracks);
  ANL4DVector  h;
  ANLPair     *zp = 0;
  while ( (zp = static_cast<ANLPair*>(zcandidates())) ) {
    ANLPair &z = *zp;
    ANLTrack &l1 = *static_cast<ANLTrack *>(z[0]);
    ANLTrack &l2 = *static_cast<ANLTrack *>(z[1]);
    Double_t lpcharge = l1.GetCharge() + l2.GetCharge();
    if (lpcharge != 0.) continue;
    Double_t zmass = z().GetMass();
    if (TMath::Abs(zmass - kMassZ) > xMll) continue;
    z.LockChildren();
    nexttrk.Reset();
    Int_t nhtrks = 0;
    while ((trkp = static_cast<ANLTrack *>(nexttrk()))) {
      ANLTrack &trk = *trkp;
      if (trk.IsLocked()) continue;
      h += trk;
      nhtrks++;
    }
    if (!nhtrks)  { z.UnlockChildren(); continue; }
    ANL4DVector qrec = qcm - z;
    Double_t rmass = qrec.GetMass();
    Double_t hmass = h.GetMass();
    if (TMath::Abs(rmass - kMassH) > xMrec) { z.UnlockChildren(); continue; }
    Double_t chi2 = TMath::Power((zmass - kMassZ)/kSigmaMz,2.)
                  + TMath::Power((rmass - kMassH)/kSigmaMh,2.);
    if (gDEBUG) {
        cerr << " M_Z = " << zmass << " M_H = " << hmass << " M_rec = " << rmass << endl;
        cerr << " zp  = " << (void*)zp
             << " hp  = " << (void*)&h << endl;
        cerr << " z[0] = " << (void*)z[0]
             << " z[1] = " << (void*)z[1] << endl;
	cerr << " chi2 = " << chi2 << endl;
    }
    solutions.Add(new ANLPair(zp,&h,chi2));
    z.UnlockChildren();
  }

  // Cut on number of solutions.
 
  if ( !solutions.GetEntries() ) { CleanUp(&tracks); return kFALSE; }
  hStat->Fill(++selid);
  if ( gNgoods == 0 ) {
    gCutName[(Int_t)selid] << "|m_ll - m_Z| <= " << xMll << " && "
	                   << "|m_x - m_H| <= " << xMrec << ends;
  }

  // Cut on cos(theta_ZH).

  TIter nextsol(&solutions);
  ANLPair *sol;
  while ( (sol = static_cast<ANLPair *>(nextsol())) ) {
    ANL4DVector &z = *static_cast<ANL4DVector *>((*sol)[0]);
    ANL4DVector &h = *static_cast<ANL4DVector *>((*sol)[1]);
    Double_t ez = z(0);
    Double_t eh = h(0);
    hEzEh->Fill(ez,eh,1.0);
    Double_t cosz = z.CosTheta();
    Double_t cosh = h.CosTheta();
    hCoszCosh->Fill(cosz,cosh,1.);
    if (TMath::Abs(cosz) > xCoszh || TMath::Abs(cosh) > xCoszh) {
      solutions.Remove(sol);
      delete sol;
    }
  }
  if ( !solutions.GetEntries() ) { CleanUp(&tracks); return kFALSE; }
  hStat->Fill(++selid);
  if ( gNgoods == 0 ) {
    gCutName[(Int_t)selid] << "|cos(theta_z)| <= " << xCoszh << "&&"
	                      "|cos(theta_H)| <= " << xCoszh << ends;
  }

  // Cut on Acop.

  nextsol.Reset();
  while ( (sol = (ANLPair*)nextsol()) ) {
    ANL4DVector &z = *static_cast<ANL4DVector *>((*sol)[0]);
    ANL4DVector &h = *static_cast<ANL4DVector *>((*sol)[1]);
    Double_t acop = z.Acop(h);
    hAcop->Fill(acop);
    if (acop > xAcop) {
      solutions.Remove(sol);
      delete sol;
    }
  }
  if ( !solutions.GetEntries() ) { CleanUp(&tracks); return kFALSE; }
  hStat->Fill(++selid);
  if ( gNgoods == 0 ) {
    gCutName[(Int_t)selid] << "Acop <= " << xAcop << ends;
  }

  // End of event selection
  
  gNgoods++;

  cerr << "------------------------------------------" << endl
       << "Event " << gJSF->GetEventNumber()
       << ": Number of solutions = " << solutions.GetEntries() << endl
       << "------------------------------------------" << endl;
  
  // Sort the solutions in the ascending order of chi2 values.

  solutions.Sort();

  // Now store this in ZH2LXAnalysisBuf.

  a.fNtracks	= fNtracks;
  a.fEvis	= fEvis;
  a.fPt		= fPt;
  a.fPl		= fPl;

  // Hists and plots for selected events.

  hNsols->Fill(solutions.GetEntries());
  hEvisPl->Fill(fEvis,fPl,1.);

  nextsol.Reset();
  Int_t nsols = 0;
  while ( (sol = static_cast<ANLPair *>(nextsol())) ) {
    if ( nsols++ ) break;
    ANLPair     &z = *static_cast<ANLPair     *>((*sol)[0]);
    ANL4DVector &h = *static_cast<ANL4DVector *>((*sol)[1]);
    Double_t chi2   = sol->GetQuality();
    Double_t zmass = z.GetMass();
    hMassz->Fill(zmass);
    Double_t hmass = h.GetMass();
    hMassh->Fill(hmass);
    hChi2->Fill(chi2);
    ANL4DVector qrec = qcm - z;
    Double_t rmass = qrec.GetMass();
    hMrec->Fill(rmass);
    Double_t cosz = z.CosTheta();
    hCosz->Fill(cosz);
    ANLTrack *lmp = static_cast<ANLTrack *>(z[0]);
    ANLTrack *lpp = static_cast<ANLTrack *>(z[1]);
    if (lmp->GetCharge() > 0.) {
       ANLTrack *tempp = lmp;
       lmp = lpp;
       lpp = tempp;
    }
    ANL4DVector &lm = *lmp;
    ANL3DVector ev3(0.,0.,1.);
    ANL3DVector evz   = z.Get3D().Unit();
    ANL3DVector evy   = (ev3^evz).Unit();
    ANL3DVector evx   = evy^evz;
    ANL3DVector evl   = lm.Get3D().Unit();
    ANL3DVector evlt  = (evl - (evl*evz)*evz).Unit();
    Double_t x = evlt*evx;
    Double_t y = evlt*evy;
    Double_t phid = TMath::ATan2(y,x);
    if (phid < 0.) phid += k2Pi;
    hPhid->Fill(phid);
  }

  // Clean up

  CleanUp(&solutions);
  CleanUp(&tracks);

  last->cd();
  return kTRUE;
}
// *****End of Process()***** //


// *****Terminate()***** //
Bool_t ZH2LXAnalysis::Terminate()
{
  // This function is called at the end of job.
  //--
  // Print out cut statistics
  //--
  cerr << endl
       << "  =============" << endl
       << "   Cut Summary " << endl
       << "  =============" << endl
       << endl
       << "  -----------------------------------------------------------" << endl
       << "   ID   No.Events    Cut Description                         " << endl
       << "  -----------------------------------------------------------" << endl;
  for (int id=0; id<kMaxCuts && gCutName[id].str().data()[0]; id++) {
    cerr << "  " << setw( 3) << id
         << "  " << setw(10) << static_cast<int>(hStat->GetBinContent(id+1))
         << "  : " << gCutName[id].str().data() << endl;
  }
  cerr << "  -----------------------------------------------------------" << endl;
  //--
  // That's it!
  //--
  return 0;
}
