#ifndef ZH2LXAnalysis_H
#define ZH2LXAnalysis_H

//**********************************************************************
//* ======================
//*  ZH2LXAnalysis Classes
//* ======================
//*
//* (Description)
//*    Analysis class for e+e- -> ZH -> 2l + X
//* (Requires)
//*	library Anlib (from physsim of K. Fujii)
//*	library ZHStudy (also in physsim libraries)
//* (Provides)
//*	class ZH2LXAnalysis
//*	class ZH2LXAnalysisBuf
//* (Usage)
//*	...
//* (Update Record)
//*	23 June 2015	K.Fujii		Original version.
//**********************************************************************

#include "TROOT.h"
#include "TFile.h"
#include "TCanvas.h"
#include "TH1.h"
#include "TH2.h"
#include "JSFSteer.h"
#include "JSFModule.h"
#include "JSFSIMDST.h"
#include "Anlib.h"

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//  ---------------------
//  ZH2LXAnalysisBuf Class
//  ---------------------
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class ZH2LXAnalysis;

class ZH2LXAnalysisBuf : public JSFEventBuf{
friend class ZH2LXAnalysis;
private:
  Int_t		fNtracks;	// track multiplicity
  Int_t         fNlptracks;     // isolated lepton multiplicity
  Double_t	fEvis;		// visible energy (GeV)
  Double_t	fPt;		// Pt
  Double_t	fPl;		// Pl

public:
  ZH2LXAnalysisBuf(const char *name="ZH2LXAnalysisBuf",
                  const char *title="ZH 2lX Data", ZH2LXAnalysis *mod=0);
  ZH2LXAnalysisBuf(ZH2LXAnalysis *mod, const char *name="ZH2LXAnalysisBuf",
                  const char *title="ZH 2lX Data");
  virtual ~ZH2LXAnalysisBuf();

  inline Int_t    GetNtracks()	const {return fNtracks; }
  inline Double_t GetEvis()	const {return fEvis; }
  inline Double_t GetPt()	const {return fPt; }
  inline Double_t GetPl()	const {return fPl; }

  ClassDef(ZH2LXAnalysisBuf, 2) // ZH2LXAnalysis Buffer Example
};

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//  ------------------
//  ZH2LXAnalysis Class
//  ------------------
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class ZH2LXAnalysis : public JSFModule {
private:
  Int_t    xNtracks;	// No. of tracks
  Double_t xEtrack;	// track energy
  Double_t xEvis;	// minimum visible energy
  Double_t xPt;		// Pt minimum
  Double_t xPl;		// Pl maximum
  Double_t xElepton;    // Elepton mimimum
  Double_t xCosCone;    // cos(theta_cone)
  Double_t xEcone;      // Ecome maximum
  Double_t xEjet;	// E_jet minimum
  Double_t xCoszh;	// |cos(theta_z)| and |cos(theta_h)| maximum
  Double_t xMll;	// |m_ll - m_Z| maximum
  Double_t xMrec;	// |m_rec - m_H| maximum
  Double_t xAcop;	// Acoplanarity maximum

public:
  TCanvas *cHist;
  TH1F *hStat;
  TH1F *hNtracks;
  TH1F *hEvis;
  TH1F *hPt;
  TH1F *hNlptracks;
  TH1F *hNsols;
  TH1F *hChi2;
  TH2F *hEvisPl;
  TH2F *hEzEh;
  TH2F *hCoszCosh;
  TH1F *hAcop;
  TH1F *hMassh;
  TH1F *hMassz;
  TH1F *hMrec;
  TH1F *hCosz;
  TH1F *hPhid;

public:
  ZH2LXAnalysis() : JSFModule("ZH2LXAnalysis","ZH2LXAnalysis Example") {}
  ZH2LXAnalysis(const Char_t *name, const Char_t *title);
  virtual ~ZH2LXAnalysis();
  
  void CleanUp(TObjArray *objs);

  inline void SetNtrackCut (Int_t    x) { xNtracks = x; }
  inline void SetEtrackCut (Double_t x) { xEtrack = x; }
  inline void SetEvisCut   (Double_t x) { xEvis = x; }
  inline void SetPtCut     (Double_t x) { xPt = x; }
  inline void SetPlCut     (Double_t x) { xPl = x; }
  inline void SetElpCut    (Double_t x) { xElepton = x; }
  inline void SetConeAngle (Double_t x) 
                          { xCosCone = TMath::Cos(TMath::Pi()*x/180.); }
  inline void SetEconeCut  (Double_t x) { xEcone = x; }
  inline void SetCoszhCut  (Double_t x)	{ xCoszh = x; }
  inline void SetMllCut    (Double_t x) { xMll   = x; }
  inline void SetMrecCut   (Double_t x) { xMrec  = x; }
  inline void SetAcopCut   (Double_t x) { xAcop  = x; }

  Bool_t Initialize();
  Bool_t Process(Int_t evt);
  Bool_t Terminate();
  void DrawHist();

  ClassDef(ZH2LXAnalysis, 2)	// ZH2LXAnalysis Example
};

#endif
