#ifndef __S4BabyHIT__
#define __S4BabyHIT__

//*************************************************************************
//* ---------------------
//* S4BabyHit Class
//* ---------------------
//* (Description)
//*	Class for a single Baby hit.
//* (Requires)
//*	TObject
//* (Provides)
//*	S4BabyHit		: TObject containing single hit data
//* (Update Record)
//*	2002/08/28  R.Kuboshima
//*************************************************************************

#include <TMath.h>
#include "S4BabyHitWire.h"
#include "S4BabyExactHit.h"
#include "S4VHit.h"

#if 0
class S4BabyHit : public U4CDCHit, public S4VHit {
#else
class S4BabyHit : public U4CDCHit {
#endif
friend class S4BabyHitWire;

public:
  S4BabyHit(S4BabyExactHit * exacthit);
  
  //==========================================================
  // --- ATTENTION ------
  // Both copy constructor and operator= don't set "this" pointer
  // to their ExactHit's Smeared Hit Pointer.
  //
     inline       S4BabyHit(const S4BabyHit & right);
     inline const S4BabyHit& operator=(const S4BabyHit& right);
  //==========================================================
  
  virtual      ~S4BabyHit();

  inline Int_t		GetTrackID       () const;
  inline Int_t          GetLayerNumber       () const;
  inline Int_t		GetCellNumber        () const;
  inline Int_t		GetWireNumber        () const;
  inline Double_t 	Xwire                () const;
  inline Double_t 	Ywire                () const;
  inline Double_t 	Zwire                () const;
  inline Double_t 	GetEnergyDeposit     () const;
  inline Int_t 		GetDriftRegionNumber () const;
  inline S4BabyHit     *GetLastHitPtr        () const;
   
  inline S4BabyExactHit* GetExactHitPtr       () const;
  inline Double_t	GetDistanceToPrevious() const;
  
  inline         void	SetDistanceToPrevious( Double_t d );  
         static  void   SetSortBy(Char_t *opt = "DriftLength");
  inline         void   SetLastHitPtr(S4BabyHit *lasthitp);
  inline virtual void   SetBit(UInt_t f, Bool_t set) 
  { 
	  TObject::SetBit(f,set); 
	  if (f & S4BabyHitWire::kEModifiedByDeltaRay) Update();
  }
  inline virtual void    SetBit(UInt_t f) 
  { 
	  TObject::SetBit(f); 
	  if (f & S4BabyHitWire::kEModifiedByDeltaRay) Update();
  }
  
  inline Bool_t	IsSortable() const;
  inline Int_t	Compare(const TObject * obj) const;

  

  void 		DebugPrint(const Char_t * opt = "Brief");

private:
  void		Update();
  void		CalculatePosition();

private:

  static Int_t 	         fS4BabyHitSortBy;	//
  static Double_t        fDXpar[4];		//
         S4BabyExactHit *fExactHitPtr;	        // Pointer to Exact Hit 
         Double_t	 fDistanceToPrevious;   // 
	 S4BabyHit      *fLastHitPtr;           //

  ClassDef(S4BabyHit, 1) 	        // Baby Anode Hit
  
};



//=====================================================================
//---------------------------------
// inline function for S4BabyHit
//---------------------------------

inline S4BabyHit::S4BabyHit(const S4BabyHit& right)
{
	fExactHitPtr         = right.fExactHitPtr;
	fDistanceToPrevious = right.fDistanceToPrevious;
	
	// !!! Don't set this to S4BabyExactHit::SmearedHitPointer !!! 
}

inline const S4BabyHit&
        S4BabyHit::operator= (const S4BabyHit& right)
{            
	fExactHitPtr         = right.fExactHitPtr;
	fDistanceToPrevious = right.fDistanceToPrevious;
	
	// !!! Don't set this to S4BabyExactHit::SmearedHitPointer !!! 
		            
  	return *this;
}

//===========================================================
// Get method -----------------------------------------------

inline  Int_t S4BabyHit::GetTrackID () const
{
	if (fExactHitPtr) {
		return fExactHitPtr->GetTrackID();
	} else {
		std::cerr << "S4BabyHit:: fExactHitPtr does not exist!" << std::endl;
		return 0;
	}
}

inline  Int_t S4BabyHit::GetLayerNumber () const
{
        if (fWirePtr) {
                return fWirePtr->GetLayerNumber();
        } else {
                return fExactHitPtr->GetLayerNumber();
        }
}

inline  Int_t S4BabyHit::GetCellNumber () const
{
	if (fWirePtr) {
		return fWirePtr->GetCellNumber();
	} else {
		return fExactHitPtr->GetCellNumber();
	}
}

inline  Int_t S4BabyHit::GetWireNumber () const
{
	if (fWirePtr) {
		return fWirePtr->GetWireNumber();
	} else {
		return fExactHitPtr->GetWireNumber();
	}
}

inline  Double_t S4BabyHit::Xwire () const
{
	if (fWirePtr) {
		return fWirePtr->X();
	} else {
		return fExactHitPtr->GetWireX();
	}
}

inline  Double_t S4BabyHit::Ywire () const
{
	if (fWirePtr) {
		return fWirePtr->Y();
	} else {
		return fExactHitPtr->GetWireY();
	}
}

inline  Double_t S4BabyHit::GetEnergyDeposit () const
{
	if (fExactHitPtr) {
		return fExactHitPtr->GetEnergyDeposit();
	} else {
		std::cerr << "S4BabyHit:: fExactHitPtr does not exist!" << std::endl;
		return 0;
	}
}

inline  Int_t S4BabyHit::GetDriftRegionNumber () const
{
	if (fWirePtr) {
		return GetDriftRegionNumber();
	} else {
		return fExactHitPtr->GetDriftRegionNumber();
	}
}

inline  S4BabyExactHit* S4BabyHit::GetExactHitPtr() const
{
	return fExactHitPtr;
}

inline  Double_t S4BabyHit::GetDistanceToPrevious() const
{
	return fDistanceToPrevious; 
}

inline  S4BabyHit *S4BabyHit::GetLastHitPtr() const
{
	return fLastHitPtr;
}

//===========================================================
// Set method -----------------------------------------------

inline  void S4BabyHit::SetDistanceToPrevious (Double_t d) 
{
	fDistanceToPrevious = d;
}

inline  void S4BabyHit::SetLastHitPtr (S4BabyHit *lasthitp)
{
	fLastHitPtr = lasthitp;
}

//===========================================================
// Other method -----------------------------------------------

inline  Bool_t S4BabyHit::IsSortable () const
{
	return kTRUE;
}

inline  Int_t S4BabyHit::Compare (const TObject *obj) const
{

  switch (fS4BabyHitSortBy) {
     case 0: { 
   	if (GetTrackID() < ((S4BabyHit *)obj)->GetTrackID()) return -1;
   	else if (GetTrackID() > ((S4BabyHit *)obj)->GetTrackID()) return 1;
  	else return 0;
  	break; }
     case 1: {
     	Double_t tof  = GetExactHitPtr()->GetTOF();
     	Double_t tofo = ((S4BabyHit *)obj)->GetExactHitPtr()->GetTOF();
   	if ( tof < tofo ) return -1;
   	else if ( tof > tofo ) return 1;
  	else return 0;
  	break; }
     case 2: {
     	Double_t edep  = GetExactHitPtr()->GetEnergyDeposit();
     	Double_t edepo = ((S4BabyHit *)obj)->GetExactHitPtr()->GetEnergyDeposit();
   	if ( edep < edepo ) return -1;
   	else if ( edep > edepo ) return 1;
  	else return 0; 
  	break;}
     case 3: {
	Double_t driftlen  = fXdrift;
	Double_t driftleno = ((S4BabyHit *)obj)->fXdrift;
   	if (driftlen < driftleno) return -1;
   	else if (driftlen > driftleno) return 1;
   	else return 0;
  	break; }
     case 4: {
     	Int_t cellno  = GetCellNumber();
     	Int_t cellnoo = ((S4BabyHit *)obj)->GetCellNumber();
   	if (cellno < cellnoo ) return -1;
   	else if (cellno > cellnoo) return 1;
  	else return 0;
  	break; }
     case 5: {
        Int_t drNumber  = GetDriftRegionNumber();
        Int_t drNumbero	= ((S4BabyHit *)obj)->GetDriftRegionNumber();
   	if (drNumber < drNumbero) return -1;
   	else if (drNumber > drNumbero) return 1;
  	else return 0;
  	break;}
     case 6: {
	Double_t driftlen  = TMath::Abs(GetExactHitPtr()->GetDriftLen());
	Double_t driftleno = TMath::Abs(((S4BabyHit *)obj)->GetExactHitPtr()->GetDriftLen());
   	if (driftlen < driftleno) return -1;
   	else if (driftlen > driftleno) return 1;
   	else return 0;
  	break; }
     case 7: {
	Double_t z  = (GetExactHitPtr()->GetPosition()).Z();
	Double_t zo = (((S4BabyHit *)obj)->GetExactHitPtr()->GetPosition()).Z();
   	if (z < zo) return -1;
   	else if (z > zo) return 1;
   	else return 0;
  	break; }
     default: {
	Double_t driftlen  = fXdrift;
	Double_t driftleno = ((S4BabyHit *)obj)->fXdrift;
   	if (driftlen < driftleno) return -1;
   	else if (driftlen > driftleno) return 1;
   	else return 0;
  	break; }
  }
}



#endif

