//*****************************************************************************
//* ==============
//* S4CALHit
//* ==============
//*   METIS (smeared) CAL hits.
//*   Mother class: S4VHit
//*
//* (Update Record)
//*   2004.07.30  A.Sanchez  Primitive version.
//*****************************************************************************
#include "S4CALHit.h"
#include "S4CALExactHit.h"
#include "S4CALPreHit.h"
#include "S4CALTypeDef.h"
#include "S4CALAddress.h"

#include "TVirtualPad.h"
#include "TVirtualViewer3D.h"
#include "TPolyMarker3D.h"
#include "TMarker3DBox.h"
#include "TViewerX3D.h"
#include "JSFSteer.h"
#if ROOT_VERSION_CODE >= ROOT_VERSION(5,15,0)
#include "TMath.h"
#endif

#include <iostream>

//#define __DEBUG__

ClassImp(S4CALHit);

//==================================================================
//* constructor ----------------------------------------------------

S4CALHit::S4CALHit()
  : S4VHit(0,0,TVector3(0,0,0),0,0),
    fCALType(0), fClusterType(0), fCellID(0),
    fCellPosition(0.,0.,0.), fIsEM(kFALSE)
{
}

S4CALHit::S4CALHit(Int_t caltype, Double_t edep, Double_t tof,
		   TVector3 pos, Int_t clustype)
  : S4VHit(edep, tof, pos, 0),
    fCALType(caltype), fClusterType(clustype),
    fCellID(0), fCellPosition(0,0,0), fIsEM(kFALSE)
{

}

S4CALHit::S4CALHit(S4CALExactHit *h)
         : S4VHit(h->GetEnergyDeposit(),       // Edep
                  h->GetTOF(),                 // TOF
#if 0
                  h->GetCellPosition(),        // hit cell position
#else
                  h->GetPosition(),            // hit position
#endif
                  0,                           // address
                  h),                          // ptr to exact hit
           fCALType(0),
           fClusterType(0),
	   fCellID(h->GetCellID()),            // cellID
	   fCellPosition(h->GetCellPosition()),// cell center positon
	   fIsEM(h->IsEM())                    // is EM ?? 

{
   if (h->IsEM()) {
      fCALType |= h->IsBarrel() ? kgBaEm : kgEcEm;
   } else {
      fCALType |= h->IsBarrel() ? kgBaHd : kgEcHd;
   }
   h->SetSmearedHitPtr(this); 
#ifdef __DEBUG__
   using namespace std;
   cerr << " pHitID = " << h->GetPreHitID()
        << " posHID = " << h->GetPreHitPtr()->GetPostHitID()
        << " Edep = " << GetEnergyDeposit()
        << " TOF = " << GetTOF()
        << endl;
#endif
}

//==================================================================
//* destructor -----------------------------------------------------

S4CALHit::~S4CALHit()
{
}

//==================================================================
//* Compare ------------------------------------------------------

Int_t S4CALHit::Compare(const TObject* obj) const
{

  switch(GetSortBy()) {
  case kEnergy:
    if ( fEnergyDeposit < ((S4CALHit*)obj)->GetEnergyDeposit() ) return -1;
    if ( fEnergyDeposit > ((S4CALHit*)obj)->GetEnergyDeposit() ) return 1;
    return 0;
    break;
  case kTof:
    if ( fTOF < ((S4CALHit*)obj)->GetTOF() ) return -1;
    if ( fTOF > ((S4CALHit*)obj)->GetTOF() ) return 1;
    return 0;
    break;
  case kPreHitID:
    if (static_cast<S4CALExactHit *>(const_cast<S4CALHit *>(this)->GetExactHitPtr())
                ->GetPreHitID() 
      < static_cast<S4CALExactHit *>(((S4CALHit*)obj)->GetExactHitPtr())
                ->GetPreHitID() ) return -1;
    if (static_cast<S4CALExactHit *>(const_cast<S4CALHit *>(this)->GetExactHitPtr())
                ->GetPreHitID()
      > static_cast<S4CALExactHit *>(((S4CALHit*)obj)->GetExactHitPtr())
                ->GetPreHitID() ) return +1;
    break;
  case kPostHitID:
    if (static_cast<S4CALExactHit *>(const_cast<S4CALHit *>(this)->GetExactHitPtr())
                ->GetPreHitPtr()->GetPostHitID() 
     <  static_cast<S4CALExactHit *>(((S4CALHit*)obj)->GetExactHitPtr())
                ->GetPreHitPtr()->GetPostHitID()) return -1;
    if (static_cast<S4CALExactHit *>(const_cast<S4CALHit *>(this)->GetExactHitPtr())
                ->GetPreHitPtr()->GetPostHitID() 
     >  static_cast<S4CALExactHit *>(((S4CALHit*)obj)->GetExactHitPtr())
                ->GetPreHitPtr()->GetPostHitID()) return +1;
    break;

  case kPosition:
    if ( fCellPosition.Mag() < (((S4CALHit*)obj)->GetCellPosition()).Mag() )
      return -1;
    if ( fCellPosition.Mag() > (((S4CALHit*)obj)->GetCellPosition()).Mag() )
      return 1;
    
    return 0;
    
    break;
  }
  return 0;
}

//==================================================================
//* DebugPrint -----------------------------------------------------

void S4CALHit::DebugPrint(const Char_t *) 
{
   S4CALExactHit *ehp = static_cast<S4CALExactHit *>(GetExactHitPtr());

   using namespace std;
   cerr << " -- S4CALHit::DebugPrint ------------------------ " << endl
        << "  htPtr = " << this                                 << endl
        << "  preID = " << ehp->GetPreHitID()                   << endl
        << "  posID = " << ehp->GetPreHitPtr()->GetPostHitID()  << endl
        << "  Xhit  = " << "(" << GetHitPosition().X() << ", "
                               << GetHitPosition().Y() << ", "
                               << GetHitPosition().Z() << ")"   << endl
        << "  Edep  = " << GetEnergyDeposit()                   << endl
        << "  pdgc  = " << ehp->GetPDGCode()                    << endl
        << " ------------------------------------------------ " << endl;
}

//=====================================================================
//* Draw --------------------------------------------------------------

void S4CALHit::Draw(Int_t color, const Char_t *opt)
{
  if (!gPad) return;

  TString option = opt;
  option.ToLower();
  gPad->cd();
  
  //TObjArray   *eharray  = GetExactHitsPtr();
  S4CALExactHit *ehitp = (S4CALExactHit*) GetExactHitPtr();    
  
  Bool_t isCLX = gJSF->Env()->GetValue( "JSFJ4.HasCLX", 1 );
  TMarker3DBox* box = 0;
  Int_t col = 0;
  
  if ( isCLX ) { // CLX geometry
    
    TVector3 pos   = ehitp -> GetCellPosition();

    static Double_t emabsthick = S4CALAddress::GetEMAbsLayerThickness();
    static Double_t emactthick = S4CALAddress::GetEMActiveLayerThickness();
    static Double_t hdabsthick = S4CALAddress::GetHDAbsLayerThickness();
    static Double_t hdactthick = S4CALAddress::GetHDActiveLayerThickness();

    Double_t cellsizex = 0.5*gJSF->Env()->GetValue("J4CLX.Cell.Size", 1.0); // the half length of cell size
    Double_t cellsizey = 0.5*gJSF->Env()->GetValue("J4CLX.Cell.Size", 1.0); // the half length of cell size
    Double_t cellthick = (ehitp->IsEM()) ? (emactthick+emabsthick+0.1)/2. 
                                         : (hdactthick+hdabsthick+0.1)/2.; // the half length of cell thickness
    
    Double_t theta = (ehitp->IsBarrel()) ? 90. : 0.;
    Double_t phi   = (Int_t)( ( TMath::ATan2(pos.Y(), pos.X())*180./TMath::Pi() + 15.) / 30.) * 30.;

    //    Double_t energy = GetEnergyDeposit();

    box = new TMarker3DBox( pos.X(), pos.Y(), pos.Z(),
			    cellsizex, cellsizey, cellthick,
			    theta, phi);
    col = color;
    
  } else { // CAL geometry
    
    TVector3 pos   = ehitp -> GetCellPosition();
    Double_t rho   = TMath::Sqrt(pos.X()*pos.X() + pos.Y()*pos.Y());
    Double_t theta = TMath::ATan2(rho, pos.Z())*180./TMath::Pi();
    Double_t phi   = TMath::ATan2(pos.Y(), pos.X())*180./TMath::Pi();
    
    static Double_t emabsthick   = gJSF->Env()->GetValue("J4CAL.EM.AbsLayerThickness",0.4);
    static Double_t emactthick   = gJSF->Env()->GetValue("J4CAL.EM.ActiveLayerThickness",0.1);
    static Double_t hdabsthick   = gJSF->Env()->GetValue("J4CAL.HD.AbsLayerThickness",0.8);
    static Double_t hdactthick   = gJSF->Env()->GetValue("J4CAL.HD.ActiveLayerThickness",0.2);
    static Double_t barcellsize  = gJSF->Env()->GetValue("J4CAL.BarrelTower.FrontSize",12.);
    static Int_t    nemminicone  = gJSF->Env()->GetValue("J4CAL.EM.MiniCone.NClones",3);
    static Int_t    nemminitower = gJSF->Env()->GetValue("J4CAL.EM.MiniTower.NClones",3);
    static Int_t    nhdminicone  = gJSF->Env()->GetValue("J4CAL.HD.MiniCone.NClones",1);
    static Int_t    nhdminitower = gJSF->Env()->GetValue("J4CAL.HD.MiniTower.NClones",1);
    
    Double_t cellsizex;  // the half length of cell size
    Double_t cellsizey;  // the half length of cell size
    Double_t cellthick;  // the half length of cell thickness
    
    ehitp->IsEM() ?
      cellsizex = barcellsize/(2.*nemminicone):
      cellsizex = barcellsize/(2.*nhdminicone);
    ehitp->IsEM() ?
      cellsizey = barcellsize/(2.*nemminitower):
      cellsizey = barcellsize/(2.*nhdminitower);
    
    ehitp->IsEM() ?
      cellthick  = (emactthick+emabsthick)/2.:
      cellthick  = (hdactthick+hdabsthick)/2.;
    
#if 0
    Double_t energy = GetEnergyDeposit();
    if(option == "e"){
      //cellsizex *= energy/(1.e-2);
      //cellsizey *= energy/(1.e-2);
      cellthick *= energy/(5.e-4);
    }
#endif
    
    box = new TMarker3DBox(pos.X(), pos.Y(), pos.Z(),
			   cellsizex, cellsizey, cellthick,
			   theta, phi);

    col = color;
  
#if 0
  if(option == "c"){
    
    energy *= 1.e+3; // GeV -> MeV
    
    col = 1;                 // black

    if(ehitp->IsEM() ? energy > 0.1 : energy > 0.2) col = 7; // light blue
    if(ehitp->IsEM() ? energy > 0.3 : energy > 0.6) col = 4; // blue
    if(ehitp->IsEM() ? energy > 0.5 : energy > 1.0) col = 3; // green
    if(ehitp->IsEM() ? energy > 0.7 : energy > 1.4) col = 5; // yellow
    if(ehitp->IsEM() ? energy > 0.9 : energy > 1.8) col = 2; // red
    
  }
#endif
  
  } // End of CLX/CAL flag check

  //box -> SetLineWidth(4);
  //box -> SetFillStyle(4100);
  //box -> SetFillColor(col);

  box->SetLineColor(col);
  box->Draw();
  gPad->Update();
  
}
