//*****************************************************************************
//* ==============
//* S4CALHitMaker
//* ==============
//*   METIS module to make (smeared) CAL hits.
//*   Mother class: S4VHit
//*
//* (Update Record)
//*   2004.07.31  A.Sanchez  Primitive version.
//*****************************************************************************
#include "S4CALHitMaker.h"
#include "S4CALMCTruthBuf.h"
#include "S4CALExactHit.h"
#include "S4CALHit.h"
#include "S4CALPreHit.h"
#include "S4CALAddress.h"
#include "JSFSteer.h"

#include "TRandom.h"

#include <map>
#include <utility>
#include <iostream>
#include <vector>
//#include "S4CALTypeDef.h"

ClassImp(S4CALHitBuf);
ClassImp(S4CALHitMaker);

Int_t    S4CALHitBuf::gfScheme = 0;
Double_t S4CALHitBuf::fgDT     = 1.3;  // nsec

//-------------------------------------------------------------------------
// S4CALHitMaker Buffer
//-------------------------------------------------------------------------

S4CALHitBuf::S4CALHitBuf(S4CALHitMaker* hitmaker, const Char_t *name, const Char_t *title, Int_t bufsize)
  : JSFEventBuf(name, title, hitmaker), fThreshold(0),
    fOutputBufPtr(0), fBufSize(bufsize)
{
}


S4CALHitBuf::~S4CALHitBuf()
{
   if (fOutputBufPtr) fOutputBufPtr->Delete();
   delete fOutputBufPtr;
}

Bool_t S4CALHitBuf::MakeHits(TObjArray *hits)
{
  if (!hits) return kTRUE;
  if (!fOutputBufPtr) fOutputBufPtr = new TObjArray(fBufSize);
  if (fOutputBufPtr->GetEntries() > 0) fOutputBufPtr->Delete();
  fOutputBufPtr->SetOwner();

  if (!gfScheme) { // realistic (1 smeared hit consists of 1 or more exact hits)
    // loop over exact hits and merge those with the same cellID
    MergeHits( hits );
  } else { // idealistic (1 smeared hit corresponds to 1 exact hit)
    TIter next(hits);
    S4CALExactHit* anExactHit = 0;
    while (( anExactHit = static_cast<S4CALExactHit*>(next()))) {
      if (anExactHit->GetEnergyDeposit() > 0.) {
	AddHit(new S4CALHit(anExactHit));
      }
    }
  }

  // smear hits and remove hits below the energy threshold
  SmearHits();
  ApplyThreshold();

  return kTRUE;
}


//============================================================================
// MergeHits -----------------------------------------------------------------
void S4CALHitBuf::MergeHits( TObjArray* hits )
{
  JSFEnv *env = gJSF->Env();

  Bool_t isCLX = env->GetValue("JSFJ4.HasCLX",1);

  if ( isCLX ) { // CLX geometry case

    S4CALExactHit::SetSortBy( "CellID" ); // sort by Cell ID
    hits->UnSort();
    hits->Sort();

    std::multimap<Int_t, S4CALExactHit*> eHitsMultiMap;
    std::map<Int_t, S4CALExactHit*> eHitsMap;

    S4CALExactHit* anExactHit = 0;
    S4CALHit*      calHitPtr  = 0;
    TIter next( hits );
    
    //* Put exactHits into multimap
    while ( anExactHit = static_cast<S4CALExactHit*>( next() ) ) {
      Int_t cellID = anExactHit->GetCellID();
      eHitsMultiMap.insert( std::make_pair( cellID, anExactHit ) );
      eHitsMap.insert( std::make_pair( cellID, anExactHit ) );
    }
    
    typedef std::map<Int_t, S4CALExactHit*>::const_iterator eHitItr;
    typedef std::multimap<Int_t, S4CALExactHit*>::const_iterator eMultiHitItr;

    //* Loop over cellID
    for ( eHitItr itr = eHitsMap.begin(); itr != eHitsMap.end(); ++itr ) {
      Int_t cellID = itr->first;

      typedef std::multimap<Int_t, S4CALExactHit*>::iterator hitItr;
      std::pair< hitItr, hitItr > hitRange = eHitsMultiMap.equal_range( cellID );

      Bool_t isFirst = kTRUE;
      Double_t tof = 999999999;
      
      for ( eMultiHitItr multiItr = hitRange.first; multiItr != hitRange.second; ++multiItr ) {

	anExactHit = multiItr->second;

	if ( isFirst ) {
	  //* Set first hit cell
	  calHitPtr = new S4CALHit( anExactHit );

	} else {
	  if ( !calHitPtr ) {
	    std::cerr << "**** Error at " << __FILE__ << " calHitPtr does not exist!"
		      << std::endl;
	  }
	  //* Add exact hit and merege energy deposit to calHit
	  calHitPtr->AddExactHitPtr( anExactHit );
	  calHitPtr->AddEnergyDeposit( anExactHit->GetEnergyDeposit() );

	  //* Update TOF information in calHit
	  if ( anExactHit->GetTOF() < tof ) {
	    calHitPtr->SetTOF( anExactHit->GetTOF() );
	  }
	}
	isFirst = kFALSE;
      }
      S4CALExactHit::SetSortBy( "EnergyDeposit" ); // sort by edep (big one first)
      const_cast<TObjArray*>( calHitPtr->GetExactHitsPtr() )->UnSort();
      const_cast<TObjArray*>( calHitPtr->GetExactHitsPtr() )->Sort();
      
      AddHit( calHitPtr );
    }
        
  } else { // CAL tower goemetry case
  
    Int_t nMergeEMMiniCones  = env->GetValue("S4CAL.EM.MiniCone.MergeCells",1);
    Int_t nMergeEMMiniTowers = env->GetValue("S4CAL.EM.MiniTower.MergeCells",1);
    Int_t nMergeHDMiniCones  = env->GetValue("S4CAL.HD.MiniCone.MergeCells",1);
    Int_t nMergeHDMiniTowers = env->GetValue("S4CAL.HD.MiniTower.MergeCells",1);
    
    S4CALExactHit::SetSortBy("CellID"); // sort by Cell ID
    hits->UnSort();
    hits->Sort();
    
    TIter next(hits);
    S4CALExactHit* anExactHit = 0;
    S4CALHit*      curhitp    = 0;
    //  S4CALAddress*  add        = 0;
    
    Double_t tofmin          = 1.e20;
    Int_t    isemprev        = -999;
    Int_t    coneidprev      = -999;
    Int_t    toweridprev     = -999;
    Int_t    miniconeidprev  = -999;
    Int_t    minitoweridprev = -999;
    Int_t    layeridprev     = -999;
    
    if((S4CALAddress::GetNMiniCones(1))%nMergeEMMiniCones!=0 
       || (S4CALAddress::GetNMiniTowers(1))%nMergeEMMiniTowers!=0 
       || (S4CALAddress::GetNMiniCones(0))%nMergeHDMiniCones!=0 
       || (S4CALAddress::GetNMiniTowers(0))%nMergeHDMiniTowers!=0 ){
      std::cerr << " Make a mistake in MergeCells. Please set MergeCells to be divisible. "<< std::endl;
      nMergeEMMiniCones  = 1;
      nMergeEMMiniTowers = 1;
      nMergeHDMiniCones  = 1;
      nMergeHDMiniTowers = 1;
    }
    
    std::map<Int_t,S4CALHit*> miniconemap;
    typedef std::map<Int_t,S4CALHit*>::iterator MCI;
    
    while( (anExactHit = static_cast<S4CALExactHit*>(next())) ){
      Int_t  cellidcur       = anExactHit->GetCellID();
      Bool_t isem            = anExactHit->IsEM(); 
      Int_t  coneidcur       = S4CALAddress::GetConeID(cellidcur);
      Int_t  toweridcur      = S4CALAddress::GetTowerID(cellidcur);
      Int_t  mergeminicones  = isem ? nMergeEMMiniCones : nMergeHDMiniCones;
      Int_t  mergeminitowers = isem ? nMergeEMMiniTowers : nMergeHDMiniTowers;
      Int_t  miniconeidquot  = S4CALAddress::GetMiniConeID(cellidcur,isem)/mergeminicones;
      Int_t  minitoweridquot = S4CALAddress::GetMiniTowerID(cellidcur,isem)/mergeminitowers;
      Int_t  layeridcur      = S4CALAddress::GetLayerID(cellidcur,isem);
      
      Bool_t istowerdata     = S4CALAddress::IsTowerData();
      if(!istowerdata){
	std::cerr << " Make a mistake to lord J4CAL.TowerGeometryData. "<< std::endl;
	nMergeEMMiniCones  = 1;
	nMergeEMMiniTowers = 1;
	nMergeHDMiniCones  = 1;
	nMergeHDMiniTowers = 1;
      }
      
      MCI hMC_itr;
      
      if( coneidprev == coneidcur && toweridprev == toweridcur && layeridprev == layeridcur 
	  && isemprev == isem && minitoweridprev == minitoweridquot 
	  && (hMC_itr = miniconemap.find(miniconeidquot)) != miniconemap.end()){
	// Merge hits in same merge cell 
	
	Double_t edep = anExactHit->GetEnergyDeposit();
	Double_t tof  = anExactHit->GetTOF();
	hMC_itr->second->AddEnergyDeposit(edep);
	hMC_itr->second->AddExactHitPtr(anExactHit);
	
	if(tof < tofmin){
	  tofmin = tof;
	  hMC_itr->second->SetTOF(tof);
	}
	
	S4CALExactHit::SetSortBy("EnergyDeposit"); // sort by edep (big one first)
	const_cast<TObjArray*>(hMC_itr->second->GetExactHitsPtr())->UnSort();
	const_cast<TObjArray*>(hMC_itr->second->GetExactHitsPtr())->Sort();
	
      }else{
	
	if( coneidprev!=coneidcur || toweridprev!=toweridcur || layeridprev!=layeridcur 
	    || isemprev!=isem || minitoweridprev!=minitoweridquot)   miniconemap.clear();
	
	if((anExactHit->GetEnergyDeposit()) > 0.){
	
	  curhitp = new S4CALHit(anExactHit);
	  
	  if ( !istowerdata ) {
	    curhitp->SetHitPosition(anExactHit->GetCellPosition());
	  } else {
	    Double_t mergex  = 0;
	    Double_t mergey  = 0;
	    Double_t mergez  = 0;
	    Int_t fminicone  = mergeminicones * miniconeidquot;
	    Int_t fminitower = mergeminitowers * minitoweridquot;
	    for(Int_t i = 0; i < mergeminitowers; i++){
	      Int_t setminitower = fminitower + i;
	      Int_t cellid = S4CALAddress::GetCellID(coneidcur,toweridcur,fminicone,setminitower,layeridcur,isem);
	      //TVector3 ancellpos = S4CALAddress::GetAddressPosition(coneidcur,toweridcur,fminicone,setminitower,layeridcur,isem);
	      TVector3 ancellpos = S4CALAddress::GetAddressPosition( cellid, isem );
	      mergex += ancellpos.X();
	      mergey += ancellpos.Y();
	    }
	    for(Int_t j=0 ; j < mergeminicones ; j++){
	      Int_t setminicone = fminicone + j;
	      Int_t cellid = S4CALAddress::GetCellID(coneidcur,toweridcur,setminicone,fminitower,layeridcur,isem);	      
	      TVector3 ancellpos = S4CALAddress::GetAddressPosition( cellid, isem );
	      //TVector3 ancellpos = S4CALAddress::GetAddressPosition(coneidcur,toweridcur,setminicone,fminitower,layeridcur,isem);
	      mergez += ancellpos.Z();
	    }
	    mergex /= mergeminitowers;
	    mergey /= mergeminitowers;
	    mergez /= mergeminicones;
	    TVector3 mergepos(mergex,mergey,mergez);
	    curhitp->SetHitPosition(mergepos);
	    //	std::cerr << "S4CALHitMaker cellX= " << mergex << " cellY= " << mergey << " cellZ= " << mergez << std::endl;
	  }
	  
	  AddHit(curhitp);
	  tofmin = 1.e20;
	  
	  isemprev        = isem;
	  coneidprev      = coneidcur;
	  toweridprev     = toweridcur;
	  layeridprev     = layeridcur;
	  minitoweridprev = minitoweridquot;
	  miniconeidprev = miniconeidquot;
	  
	  miniconemap[miniconeidquot] = curhitp;
	}
      }
    }
  }
}

void S4CALHitBuf::SortHits(Int_t option)
{
  S4CALHit::SetSortBy(option);
  fOutputBufPtr->UnSort();
  fOutputBufPtr->Sort();
}

void S4CALHitBuf::SmearHits()
{

  TIter next(fOutputBufPtr);
  S4CALHit* chit = 0;

  while( (chit = (S4CALHit*)next()) ){
    chit->SetTOF(chit->GetTOF() + gRandom->Gaus(0., fgDT));
    if(chit->GetTOF() < 1.e-6) chit->SetTOF(1.e-6);
  }

}

void S4CALHitBuf::ApplyThreshold()
{

}

//============================================================================
// S4CALHitMaker
//============================================================================
S4CALHitMaker::S4CALHitMaker(const Char_t *name,
			     const Char_t *title,
			     const Char_t *opt)
  : JSFModule(name, title, opt) 
{
  fEventBuf = new S4CALHitBuf(this,"S4CALHitBuf","S4CALHits Data");
}

//============================================================================
S4CALHitMaker::~S4CALHitMaker()
{
  if(fEventBuf) delete fEventBuf;
}

//============================================================================
Bool_t S4CALHitMaker::Initialize()
{
  if(fEventBuf) delete fEventBuf;
  fEventBuf = new S4CALHitBuf(this);
  return kTRUE;
}

//============================================================================
Bool_t S4CALHitMaker::Process(Int_t ev)
{
  S4CALHitBuf     *hmbufp  = static_cast<S4CALHitBuf *>(EventBuf());
   
  S4CALMCTruth    *mctp    = (S4CALMCTruth *)gJSF->FindModule("S4CALMCTruth","quiet");
  S4CALMCTruthBuf *mctbufp = (S4CALMCTruthBuf *)mctp->EventBuf();
  hmbufp->SetInputDataBufPtr(mctbufp); 

  return hmbufp->MakeHits(mctbufp->GetExactHitsPtr());
}
