#ifndef __S4CALSMALLCLUSTERUTILS__
#define __S4CALSMALLCLUSTERUTILS__
//*************************************************************************   
//* ---------------------                      
//* S4CALSmallClusterUtils Class
//* ---------------------   
//* (Description) 
//*     represent calorimeter cluster
//* (Requires)
//*
//* (Provides)  
//*
//* (Update Record)
//*
//*
//*************************************************************************

#include "S4CALSmallCluster.h"
#include "S4HYBTrack.h"
#include "JSFSteer.h"
#include "TVector3.h"
#include "TObject.h"

#include <utility>
#include <map>
#include <vector>

class S4CALSmallClusterUtils {

friend class S4CALSmallClusterMakerBuf;

 public:
  S4CALSmallClusterUtils();
  virtual ~S4CALSmallClusterUtils();

  //=================================================================
  //* Service method
  //-----------------------------------------------------------------

 public:

  //===============================================================
  // Make clusters using LDC group's method.
  // If isem = kTRUE, small clusters in ECAL/HCAL will be separated.
  // Note that "Locked" hits will not be used.

  static std::vector<S4CALSmallCluster*> MakeClustersLDC(std::map<Int_t,S4CALHit*> &hitmap, Bool_t isem = kFALSE);

  //===============================================================
  // Merge small clusters in an invec by dX dT method.
  // If isem = kTRUE, small clusters in ECAL/HCAL will be separated.
  // Note that if small cluster was locked or was MIP, it will not be merged.

  static std::vector<S4CALSmallCluster*> MergeSmallClustersdXdT(std::vector<S4CALSmallCluster*> invec, Bool_t isem = kFALSE, Char_t* opt = "");
  

  //===============================================================
  // Merge small clusters in an invec by LDC small clustering method.
  // If isem = kTRUE, small clusters in ECAL/HCAL will be separated.
  // Note that if small cluster is locked or is MIP, it will not be merged.
  
  static std::vector<S4CALSmallCluster*> MergeSmallClustersLDCSCL(std::vector<S4CALSmallCluster*> invec, Bool_t isem = kFALSE);
  

  //===============================================================
  // Merge small clusters in a sclvec by LDC method.
  // If isem = kTRUE, small clusters in ECAL/HCAL will be separated.
  // Note that if small cluster is locked or is MIP, it will not be merged.

  static std::vector<S4CALSmallCluster*> MergeSmallClustersLDC(std::vector<S4CALSmallCluster*> &sclvec, Bool_t isme = kFALSE);

  
  //===============================================================
  // IL finder which return pair(IL position, S4CALHit* vector).
  // If it has no IL position, return empty vector.

  static std::pair<TVector3,std::vector<S4CALHit*> > ILFinder(std::map<Int_t,S4CALHit*> &hitmap, S4HYBTrack* trk);

  //===============================================================
  // Cheated IL position finder.
  // Return (post/pre hit track ID, PID, IL position).

  static std::map<Int_t,std::pair<Int_t,TVector3> > CheatedILFinder(std::map<Int_t,S4CALHit*> hitmap);

  //===============================================================
  // Calculate the nearest track from pos. Return value is
  // (track ID, distance).
  // Note that locked track will not be checked!

  static std::pair<Int_t,Double_t> NearestTrackIs(std::map<Int_t, S4HYBTrack*> &trkmap,
						  TVector3 pos,
						  Double_t pcut = 0.); // P cut

  //==============================================================
  // Calculate the distance from all tracks in trkmap.
  // Note that locked track will not be checked!

  static std::multimap<Double_t,S4HYBTrack*> GetTrackDistanceList(std::map<Int_t,S4HYBTrack*> &trkmap,
							     TVector3 pos,
							     Double_t pcut = 0.);
  
  //==============================================================
  // Calculate the distance from track to sampling position.
  // (sampling position shold be outside of CAL inner surface.)

  static Double_t CalcTrackDistance(S4HYBTrack* trk, const TVector3 pos);
 
  //==============================================================
  // Calculate track intersection point with cylinder.
  // The vector of return value has
  // vector[0]: track intersection point
  // vector[1]: track momentum at intersection point

  static std::vector<TVector3> GetInTrack(S4HYBTrack *track,
					  const Double_t tubeR     = 210.,
					  const Double_t tubehalfH = 270.);

  //===============================================================
  // Calculate density of energy deposit in the sphere
  // An unit of return value is (GeV/cm^3)

  static Double_t  CalcEnergyDensity(std::map<Int_t,S4CALHit*> hitmap, TVector3 pos, Double_t radius = 20.);

  //===============================================================
  // Sort hits by distance from IP (temporary!)

  static std::vector<S4CALHit*> SortHitMap(std::map<Int_t,S4CALHit*> &hitmap);


  //===============================================================
  // Setter

  //------ For MakeClustersLDC -------
  static void SetEMRCutMin(Double_t cut) { fgEMRCutMin = cut;   }
  static void SetEMRCutMax(Double_t cut) { fgEMRCutMax = cut;   }
  static void SetHDRCutMin(Double_t cut) { fgHDRCutMin = cut;   }
  static void SetHDRCutMax(Double_t cut) { fgHDRCutMax = cut;   }
  static void SetEMDCut(Double_t cut)    { fgEMDCut    = cut;   }
  static void SetHDDCut(Double_t cut)    { fgHDDCut    = cut;   }
  static void SetEMNSteps(Int_t steps)   { fgEMNSteps  = steps; }
  static void SetHDNSteps(Int_t steps)   { fgHDNSteps  = steps; }
  //----------------------------------

  //--- For MergeSmallClustersLDC ----
  static void SetEMDistCutBack(Double_t cut) { fgEMDistCutBack = cut; }
  static void SetEMDistCutFwd(Double_t cut)  { fgEMDistCutFwd  = cut; }
  static void SetHDDistCutBack(Double_t cut) { fgHDDistCutBack = cut; }
  static void SetHDDistCutFwd(Double_t cut)  { fgHDDistCutFwd  = cut; }
  static void SetNHitsCut(Int_t cut)         { fgNHitsCut      = cut; }
  static void SetMinLayCut(Int_t cut)        { fgMinLayCut     = cut; }
  //----------------------------------

  //--- For MergeSmallClustersdXdT ---
  static void SetEMdX(Double_t dx) { fgEMdX = dx; }
  static void SetHDdX(Double_t dx) { fgHDdX = dx; }
  static void SetEMdT(Double_t dt) { fgEMdT = dt; }
  static void SetHDdT(Double_t dt) { fgHDdT = dt; }
  //----------------------------------

  //--------- For ILFinder -----------
  static void SetILEMTubeR(Double_t r)           { fgILEMTubeR      = r;      }
  static void SetILHDTubeR(Double_t r)           { fgILHDTubeR      = r;      }
  static void SetILMIPCutEnergy(Double_t energy) { fgILMIPCutEnergy = energy; }
  //----------------------------------

  //=================================================================
  //* Data members
  //-----------------------------------------------------------------
  
 private:

  //------ Paramters for MakeClustersLDC() ------
  static Double_t fgEMRCutMin;
  static Double_t fgEMRCutMax;
  static Double_t fgHDRCutMin;
  static Double_t fgHDRCutMax;
  static Double_t fgEMDCut;
  static Double_t fgHDDCut;
  static Int_t    fgEMNSteps;
  static Int_t    fgHDNSteps;
  //----------------------------------------------
  
  //--- Parameters for MergeSmallClustersLDC() ---
  static Double_t fgEMDistCutBack;
  static Double_t fgEMDistCutFwd;
  static Double_t fgHDDistCutBack;
  static Double_t fgHDDistCutFwd;
  static Int_t    fgNHitsCut;
  static Int_t    fgMinLayCut;
  //----------------------------------------------

  //----- Paramters for MergeClustersdXdT() ------
  static Double_t fgEMdX;
  static Double_t fgHDdX;
  static Double_t fgEMdT;
  static Double_t fgHDdT;
  //----------------------------------------------

  //--------- Parameters for ILFinder() ----------
  static Double_t fgILEMTubeR;
  static Double_t fgILHDTubeR;
  static Double_t fgILMIPCutEnergy;
  //----------------------------------------------

  ClassDef(S4CALSmallClusterUtils, 1)
};
    
#endif
