//*************************************************************************
//* ---------------------
//* S4CDCHit Class
//* ---------------------
//* (Description)
//*	Class for a single CDC hit.
//* (Requires)
//*	TObject
//* (Provides)
//*	S4CDCHit		: TObject containing single hit data
//* (Update Record)
//*	1999/12/04  K.Fujii	Original version.
//*************************************************************************

#include "S4CDCHit.h"
#include "TRandom.h"
#include "TMath.h"
#include "U4CDCMeasLayer.h"
#include "U4CDCKalDetector.h"

#define __TMP__

#ifdef __TMP__
#define __DUMMYDXDRIFT__ 0.0085
#define __DUMMYVDRIFT__  669600      // [cm/s] at 0.9 [kV/cm]
static const Double_t kNS = 1.e-9;   // [ns]

#if 1
#define __DUMMYDZA__ 	 5.
#define __DUMMYDZS__	 5.
#else
#define __DUMMYDZA__ 	 0.0085
#define __DUMMYDZS__	 0.0085
#endif
//#define __DUMMYDZS__	 10.*__DUMMYDXDRIFT__
#endif

ClassImp(S4CDCHit)

Int_t    S4CDCHit::fgS4CDCHitSortBy = 0; 

//////////////// VERY TEMPORARY ///////////////////////////////////////

Double_t S4CDCHit::fgDXpar[] = { 3.83683e-04, 5.83973e+00, 4.04081e-05, 1.24272e-05 };
TString  S4CDCHit::fgSpatialResolutionCurve = "flat";

Double_t S4CDCHit::fgT0pedestal = 0.; // [sec]

///////////////////////////////////////////////////////////////////////

//=====================================================================
//* constructor -------------------------------------------------------

S4CDCHit::S4CDCHit(S4CDCExactHit * exacthit)
	 : fExactHitPtr(exacthit), fDistanceToPrevious(9999.),
	   fLastHitPtr(0)
{	
	fExactHitPtr->SetSmearedHitPtr(this);
}

//=====================================================================
//* destructor --------------------------------------------------------

S4CDCHit::~S4CDCHit() 
{
}
  
  
//=====================================================================
//* SetSortBy ------------------------------------------------------=--

void S4CDCHit::SetSortBy(Char_t *opt )
{
	switch (opt[2]) {
		case 'a':	// sort by "TrackNumber"
			fgS4CDCHitSortBy = 0;
			break;
		case 'm': 	// sort by "Time"
			fgS4CDCHitSortBy = 1;
			break;
		case 'e': 	// sort by "EnergyDeposit"
			fgS4CDCHitSortBy = 2;
			break;
		case 'i': 	// sort by "DriftLength"
			fgS4CDCHitSortBy = 3;
			break;
		case 'l': 	// sort by "CellID"
			fgS4CDCHitSortBy = 4;
			break;
		case 'r': 	// sort by "WireLayer"
			fgS4CDCHitSortBy = 5;
			break;
		case 'u': 	// sort by "TrueDriftLength"
			fgS4CDCHitSortBy = 6;
			break;
		case 'o': 	// sort by "ZCoordinate"
			fgS4CDCHitSortBy = 7;
			break;
		default: 	// sort by "DriftLength"
			fgS4CDCHitSortBy = 5;
	}		
}

//=====================================================================
//* DebugPrint --------------------------------------------------------

void S4CDCHit::DebugPrint(const Char_t *opt)
{
	if (!fDone) Update();
	std::cerr.setf(std::ios::right);
	std::cerr.setf(std::ios::fixed);
#if 1
        Int_t   layerno  = fExactHitPtr->GetLayerNumber();
        Int_t   wireno   = fExactHitPtr->GetWireNumber();

        Int_t           lno   = layerno * 5 + wireno;
        U4CDCMeasLayer *msPtr = dynamic_cast<U4CDCMeasLayer *>(U4CDCKalDetector::GetInstance()->GetMeasLayerPtr(lno));
        U4CDCMeasLayer &ms    = *msPtr;

        TVector3       xx     = fExactHitPtr->GetPosition();

        Int_t lr     = 0;
        Int_t cellno = -999999999;
        TKalMatrix h = ms.XvToMv(xx, lr, cellno);

        std::cerr << " (layr, wir, cell, lr) = (" 
                     << std::setw(2) << layerno << "," 
                     << std::setw(2) << wireno  << "," 
                     << std::setw(3) << cellno  << "," 
                     << std::showpos 
                     << std::setw(3) << lr      << ")"
                     << std::noshowpos 
                  << " xx = (" 
                     << std::setw(8) << std::setprecision(3) << xx.X() << "," 
                     << std::setw(8) << std::setprecision(3) << xx.Y() << "," 
                     << std::setw(8) << std::setprecision(3) << xx.Z() << ")"
                  << " (X,Y) = (" 
                     << std::setw(8) << std::setprecision(3) << fWirePtr->X() << "," 
                     << std::setw(8) << std::setprecision(3) << fWirePtr->Y() << ")"
                  << " (d, z) = (" 
                     << std::setw(6) << std::setprecision(3) << fXdrift   << "+/-"
                     << std::setw(6) << std::setprecision(4) << fDXdrift  << ", "
                     << std::setw(5) << std::setprecision(1) << fZ        << "+/-"
                     << std::setw(3) << std::setprecision(1) << fDZ       << ")" 
                  << std::endl;
#else 
	std::cerr	<< " TrackNumber = " << fExactHitPtr->GetTrackID()
		<< " PDGcode = " << fExactHitPtr->GetPDGCode()
		<< " CellID = " << fExactHitPtr->GetCellID()
		<< " WireNumber = " << fExactHitPtr->GetWireNumber()
		<< " DriftLength = " << fExactHitPtr->GetDriftLen()
		<< " TotalEnergy = " << fExactHitPtr->GetTotalEnergy()
		<< " KillHitFlag = " << (TestBit(S4CDCHitWire::kECannotSeparate) ? 1 : 0)
		<< std::endl;
#endif
		
	std::cerr.unsetf(std::ios::right);
	std::cerr.unsetf(std::ios::fixed);
}

//=====================================================================
//* Update ------------------------------------------------------------

void S4CDCHit::Update()
{
        U4CDCHit::Update();
}

//=====================================================================
//* CalculatePosition() -----------------------------------------------

void S4CDCHit::CalculatePosition()
{
	if (fWirePtr) {

		S4CDCExactHit *exacthitp = 0;
		if (fLastHitPtr && TestBit(S4CDCHitWire::kEModifiedByDeltaRay)) {
                   exacthitp = (S4CDCExactHit *)fLastHitPtr->GetExactHitPtr();
                } else {
                   exacthitp = fExactHitPtr;
                }
#if 1
		fXdrift  = TMath::Abs(exacthitp->GetDriftLen());
		fZ	 = fExactHitPtr->GetPosition().Z();
                Int_t lr = (fExactHitPtr->GetDriftLen() < 0. ? -1 : 1);
                fXdrift *= lr;
#else
                //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
                //>>>>>>>>>> very temporary treatment!!! >>>>>>>>>>>>>>>>>>
                //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

                Int_t   layerno  = exacthitp->GetLayerNumber();
                Int_t   wireno   = exacthitp->GetWireNumber();

                Int_t           lno   = layerno * 5 + wireno;
                U4CDCMeasLayer *msPtr = U4CDCKalDetector::GetInstance()->GetMeasLayerPtr(lno);
                U4CDCMeasLayer &ms    = *msPtr;

                TVector3       xx     = exacthitp->GetPosition();

                Int_t lr     = 0;
                Int_t cellno = -999999999;
                TKalMatrix h = ms.XvToMv(xx, lr, cellno);

                fXdrift = h(0,0);
                fZ      = h(1,0);

                //<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                //>>>>>>>>>> very temporary treatment!!! >>>>>>>>>>>>>>>>>>
                //<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
#endif
		if (fgSpatialResolutionCurve == "experimental") {
		   fDXdrift = TMath::Sqrt( fgDXpar[0]*TMath::Exp(-fgDXpar[1]*fXdrift)
		                         + fgDXpar[2]*fXdrift + fgDXpar[3] );
		} else if (fgSpatialResolutionCurve == "flat") {
		   fDXdrift = __DUMMYDXDRIFT__ ;		
		}

		fXdrift += gRandom->Gaus(0.,fDXdrift);

                fVdrift  = __DUMMYVDRIFT__ * kNS;

		if (fWirePtr->WhatAmI()) {
			fDZ	 = __DUMMYDZS__;
		} else {
			fDZ	 = __DUMMYDZA__;
		}
			
		fZ	+= gRandom->Gaus(0.,fDZ);

                fXdrift += fVdrift * fgT0pedestal * lr;
#if 0
                DebugPrint();
#endif
	} else {
		std::cerr 	<< "::::: Error in S4CDCHit::CalculatePosition"  << std::endl
			<< " This hit belong to no S4CDCHitWire object!" << std::endl
			<< " Append this to some S4CDCHitWire object!"   << std::endl;
	}
}
