//*****************************************************************************
//* =====================
//*  EHEHSpring
//* =====================
//*  
//* (Description)
//*    e+e- --> eH eH generator
//*    Based on example/FFbarSpring directory.
//*
//* (Update Record)
//*    2010/06/26  E.Kato	Original version.
//*
//*****************************************************************************


#include "JSFSteer.h"
#include "EHEHSpring.h"

#include <sstream>
#include <iomanip>
//#define __NODECAY__
//#define __DEBUG__
//#define __ZEROWIDTH__
//#define __PHASESPACE__
#ifdef __NODECAY__
#ifndef __ZEROWIDTH__
#define __ZEROWIDTH__
#endif
#endif

//*----------------------------------------------------------------------
//*     Numerical and Natural Constants
//*----------------------------------------------------------------------
#include "GENNumCon.h"

ClassImp(EHEHSpring)
ClassImp(EHEHSpringBuf)
ClassImp(EHEHBases)
ClassImp(ZHBoson)
ClassImp(WHBoson)
ClassImp(AHBoson)
ClassImp(EHFermion)
//ClassImp(ZBoson)

//-----------------------------------------------------------------------------
// ==============================
//  class EHEHSpring
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
EHEHSpring::EHEHSpring(const char      *name,
                       const char      *title,
                             EHEHBases *bases)
          : JSFSpring(name, title, bases)
{
  fEventBuf = new EHEHSpringBuf("EHEHSpringBuf",
                                "EHEHSpring event buffer",
                                this);
  if (!bases) { 
    SetBases(new EHEHBases);
  }
}

//_____________________________________________________________________________
// --------------------------
//  D-tor
// --------------------------
EHEHSpring::~EHEHSpring()
{
  delete fEventBuf;
  delete GetBases();
}

//_____________________________________________________________________________
// --------------------------
//  Initialize
// --------------------------
Bool_t EHEHSpring::Initialize()
{
  JSFSpring::Initialize();

  if (fFile->IsWritable()) {
    EHEHBases *bs = static_cast<EHEHBases *>(GetBases());
    TDirectory *last = gDirectory;
    fFile->cd("/conf");
    TList *dlist = gDirectory->GetListOfKeys();
    if (!dlist->FindObject("init")) gDirectory->mkdir("init");
    fFile->cd("/conf/init");
    bs->Write();
    last->cd(); 
    cerr << ">>>>>> EHEHBases written to file" << endl;
  }

  return kTRUE;
}

//-----------------------------------------------------------------------------
// ==============================
//  class EHEHSpringBuf
// ==============================
//_____________________________________________________________________________
// --------------------------
//  SetPartons
// --------------------------
Bool_t EHEHSpringBuf::SetPartons()
{
  TClonesArray &partons = *fPartons;
  EHEHBases  *bases     = static_cast<EHEHBases *>
                         (static_cast<EHEHSpring *>(Module())->GetBases());

  // ----------------------------------------------
  //  Set 4-momenta of final state particles in CM
  // ----------------------------------------------
  const Int_t kNparton = 6;
  ANL4DVector pv[kNparton];
  pv[2] = bases->fP[0];    // e- from eH-
  pv[3] = bases->fP[1];    // AH from eH-
  pv[4] = bases->fP[2];    // e+ from eH+
  pv[5] = bases->fP[3];    // AH from eH+
  pv[0] = pv[2] + pv[3];   // eH-
  pv[1] = pv[4] + pv[5];   // eH+

  // ----------------------------------------------
  //  Boost them to lab. frame
  // ----------------------------------------------

  fZBoost       = bases->GetZBoost();
  fEcmIP        = bases->GetEcmIP();
  fQ2XX         = fEcmIP*fEcmIP;
  fCosTheta     = bases->GetCosTheta();
  fPhi          = bases->GetPhi();
  fQ2X1         = bases->GetQ2X1();
  fCosThetaEm   = bases->GetCosThetaEm();
  fPhiEm        = bases->GetPhiEm();
  fQ2X2         = bases->GetQ2X2();
  fCosThetaEp   = bases->GetCosThetaEp();
  fPhiEp        = bases->GetPhiEp();
  Double_t elab = TMath::Sqrt(fEcmIP*fEcmIP + fZBoost*fZBoost);
  TVector3 boostv(0.,0.,fZBoost/elab);
  for (Int_t i=0; i<kNparton; i++) pv[i].Boost(boostv);

  // ----------------------------------------------
  //  Set final state parton infomation
  // ----------------------------------------------

  fNparton = kNparton;

  static TVector **qp = 0;
  if (!qp) {
    qp = new TVector* [kNparton];
    for (Int_t i=0; i<kNparton; i++) qp[i] = new TVector(4);
  }
  for (Int_t i=0; i<kNparton; i++) {
    TVector &q = *qp[i];
    q(0) = pv[i].E ();
    q(1) = pv[i].Px();
    q(2) = pv[i].Py();
    q(3) = pv[i].Pz();
  }
  Int_t    idx     = bases->fXFermionPtr ->GetPID(); // PDG code for eH? 
  Double_t mass    = bases->GetMass();

  Int_t    iddm    = bases->fDMBosonPtr->GetPID(); // PDG code for AH? 
  Double_t msdm    = bases->GetMassDM();
  Double_t me      = kM_e;

  Int_t    ide     = 11;                           // PDG code for e

  //                               No. PID   Mass Charge   pv     Nd 1st Mom hel col shower
  new (partons[0]) JSFSpringParton( 1,  idx , mass,   -1., *qp[0], 2, 3,  0, 0,   0,     0);
  new (partons[1]) JSFSpringParton( 2, -idx , mass,   +1., *qp[1], 2, 5,  0, 0,   0,     0);
  new (partons[2]) JSFSpringParton( 3,  ide , me  ,   -1., *qp[2], 0, 0,  1, 0,   0,     0);
  new (partons[3]) JSFSpringParton( 4,  iddm, msdm,    0., *qp[3], 0, 0,  1, 0,   0,     0);
  new (partons[4]) JSFSpringParton( 5, -ide , me  ,   +1., *qp[4], 0, 0,  2, 0,   0,     0);
  new (partons[5]) JSFSpringParton( 6,  iddm, msdm,    0., *qp[5], 0, 0,  2, 0,   0,     0);

  return kTRUE ;
}

//-----------------------------------------------------------------------------
// ==============================
//  class EHEHBases
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
EHEHBases::EHEHBases(const char *name, const char *title)
         : JSFBases   (name, title), 
	   fF         ( 580.),
	   fKappaL    (  0.5),
           fMass      ( 369.),
           fMassDM    ( 81.9),
           fMassHiggs ( 134.),
           fEcmInit   (1000.),
           fISR       ( 1),
           fIBType    ( 2),
           fBeamStr   ( 1),
           fBeamWidth (0.002),
           fPole      (0.),
           fXFermionPtr(0),
	   fWHBosonPtr( 0),
	   fZHBosonPtr( 0),
	   fDMBosonPtr( 0),
           fWBosonPtr ( 0),
           fZBosonPtr ( 0),
           fPhotonPtr ( 0),
           fGluonPtr  ( 0),
           fZBoost    (0.),
           fEcmIP     (fEcmInit),
           fQ2XX      (0.),
           fQ2X1      (0.),
           fQ2X2      (0.),
           fCosTheta  (0.),
           fPhi       (0.),
           fCosThetaEm(0.),
           fPhiEm     (0.),
           fXQ2X1     (0.),
           fCosThetaEp(0.),
           fPhiEp     (0.),
           fXQ2X2     (0.),
           fR_BW_m    (0),
           fR_BW_p    (0),
           fR_BS_m    (0),
           fR_BS_p    (0),
           fR_ISR_var (0),
           fR_ISR_side(0)
{
  //  Constructor of bases.  Default parameter should be initialized here
  //
  // --------------------------------------------
  //  Get parameters from jsf.conf, if specified
  // --------------------------------------------

  cout << "Init EHEHBases " << endl;
  
  using namespace std;
  stringstream ins(gJSF->Env()->GetValue("EHEHBases.CosthRange","-1.0 1.0"));
  ins >> fXL[0] >> fXU[0];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("EHEHBases.PhiOverPiRange","0.0 2.0"));
  ins >> fXL[1] >> fXU[1];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("EHEHBases.CosthEmRange","-1.0 1.0"));
  ins >> fXL[2] >> fXU[2];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("EHEHBases.PhiEmOverPiRange","0.0 2.0"));
  ins >> fXL[3] >> fXU[3];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("EHEHBases.CosthEpRange","-1.0 1.0"));
  ins >> fXL[4] >> fXU[4];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("EHEHBases.PhiEpOverPiRange","0.0 2.0"));
  ins >> fXL[5] >> fXU[5];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("EHEHBases.F","580.")); 	 // F [GeV]
  ins >> fF;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("EHEHBases.KappaL","0.5")); 	 // kappa_l
  ins >> fKappaL;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("EHEHBases.MassHiggs","134."));  // M_h [GeV]
  ins >> fMassHiggs;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("EHEHBases.Ecm","1000."));       // E_cm (1TeV)
  ins >> fEcmInit;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("EHEHBases.Beamstrahlung","1")); // BmStr (on)
  ins >> fBeamStr;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("EHEHBases.IBType","2")); // Beam spread type (1: Uniform, 2: Gauss)
  ins >> fIBType;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("EHEHBases.BeamWidth","0.0011")); // Beam spread (relative)
  ins >> fBeamWidth;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("EHEHBases.Bremsstrahlung","1"));// ISR (on)
  ins >> fISR;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("EHEHBases.Pole","0."));         // electron polarization
  ins >> fPole;


  // --------------------------------------------
  //  Registor random numbers
  // --------------------------------------------
  //--
  //  Initial state helicity combination
  //--
  DefineVariable(fHelCombInitial, 0., 1., 1, 1);
  DefineVariable(fHelCombFinal  , 0., 1., 1, 1);
  DefineVariable(fXQ2X1         , 0., 1., 0, 1);
  DefineVariable(fXQ2X2         , 0., 1., 0, 1);
  
  //--
  //  cos(theta) and phi
  //--
  fXL[1] = fXL[1]*TMath::Pi();
  fXU[1] = fXU[1]*TMath::Pi();
  fXL[3] = fXL[3]*TMath::Pi();
  fXU[3] = fXU[3]*TMath::Pi();
  fXL[5] = fXL[5]*TMath::Pi();
  fXU[5] = fXU[5]*TMath::Pi();

  DefineVariable(fCosTheta  , fXL[0], fXU[0], 1, 1);
  DefineVariable(fPhi       , fXL[1], fXU[1], 0, 0);
  DefineVariable(fCosThetaEm, fXL[2], fXU[2], 0, 0);
  DefineVariable(fPhiEm     , fXL[3], fXU[3], 0, 0);
  DefineVariable(fCosThetaEp, fXL[4], fXU[4], 0, 0);
  DefineVariable(fPhiEp     , fXL[5], fXU[5], 0, 0);

  //--
  //  Beamstrahlung and natural Ebeam spread
  //--
  if (fBeamStr == 1) {
    DefineVariable(fR_BW_m, 0., 1., 0, 1);
    DefineVariable(fR_BW_p, 0., 1., 0, 1);
    DefineVariable(fR_BS_m, 0., 1., 0, 1);
    DefineVariable(fR_BS_p, 0., 1., 0, 1);
  }
 
  //--
  //  ISR
  //--
  if (fISR==1) {
    DefineVariable(fR_ISR_var,  0., 1., 0, 1);
    DefineVariable(fR_ISR_side, 0., 1., 0, 1);
  }

  // --------------------------------------------
  //  Set Bases integration parameters
  // --------------------------------------------
  SetNoOfSample(40000);

  SetTuneValue (1.5);
  SetIteration1(0.05, 20);
  SetIteration2(0.05,100);

}
// --------------------------
//  D-tor
// --------------------------
EHEHBases::~EHEHBases()
{
  delete fXFermionPtr;
  delete fWHBosonPtr;
  delete fZHBosonPtr;
  delete fDMBosonPtr;
  delete fWBosonPtr;
  delete fZBosonPtr;
  delete fPhotonPtr;
  delete fGluonPtr;
}

//_____________________________________________________________________________
// --------------------------
//  Func
// --------------------------
Double_t EHEHBases::Func()
{
  //  Bases Integrand
  //

  Double_t bsWeight = 1.; // Jacobian factor
  Double_t eminus;        // E_e- after ISR and beamstrahlung
  Double_t eplus;         // E_e+ after ISR and beamstrahlung

  // --------------------------------------------
  //  Beamstrahlung
  // --------------------------------------------
  if (fBeamStr == 1) {
    bsWeight = fBM->GetWeight(fR_BW_m, fR_BW_p,
                              fR_BS_m, fR_BS_p,
                              eminus, eplus);
#if 0
    eplus   *= fBM->GetNominalEnergy();
    eminus  *= fBM->GetNominalEnergy();
#else
    eplus   *= fEcmInit/2.;
    eminus  *= fEcmInit/2.;
#endif
    fEcmIP   = 2.*TMath::Sqrt(eplus*eminus);     // ignore electron mass 
  } else {
    fEcmIP   = fEcmInit;      
    eplus    = fEcmInit/2.;
    eminus   = eplus;
  }
   
  // --------------------------------------------
  //  Initial State Radiation
  // --------------------------------------------

  static const Double_t kFisr = 2. * kAlpha0 / TMath::Pi();

  Double_t beta_isr = kFisr * (2.*(TMath::Log(fEcmIP/kM_e) - 1.));

  if (fISR == 1) {
    Double_t xisr    = TMath::Power(fR_ISR_var,1./beta_isr); // Ephoton / Ebeam 
             fEcmIP *= TMath::Sqrt(1. - xisr);                // reduced Ecm
    
    if (fR_ISR_side > 0.5) {
      eplus  *= 1. - xisr;
    } else {
      eminus *= 1. - xisr;
    }
  } // if Bremsstrahlung


  fZBoost = eminus - eplus; // P_z of the cm system after ISR and beamstrahlung

  // --------------------------------------------
  //  Select final state
  // --------------------------------------------
  Double_t weight = 1;

  // --------------------------------------------
  //  Check if reduced Ecm enough for prod.
  // --------------------------------------------
  if (fEcmIP < 2*kM_e + 2*fMassDM) {
    return 0.;
  }

  // --------------------------------------------
  //  Select helicity combination
  // --------------------------------------------
  //  Notice that spin average for e- is taken
  //  care of here
  SelectHelicities(weight);
  bsWeight *= weight;

  // --------------------------------------------
  //  Decide Q^2 of internal lines
  // --------------------------------------------
  fQ2XX = fEcmIP*fEcmIP;

  Double_t rs   = fEcmIP;
  Double_t qmin = kM_e + fMassDM;
  Double_t qmax = rs - kM_e - fMassDM;
#ifndef __ZEROWIDTH__
  fQ2X1 = fXFermionPtr->GetQ2BW(qmin, qmax, fXQ2X1, weight);
  Double_t qx1 = TMath::Sqrt(fQ2X1);
  bsWeight *= weight;
  qmax = rs - qx1;
  fQ2X2 = fXFermionPtr->GetQ2BW(qmin, qmax, fXQ2X2, weight);
  bsWeight *= weight;
#else
  fQ2X1 = TMath::Power(fXFermionPtr->GetMass(),2);
  fQ2X2 = TMath::Power(fXFermionPtr->GetMass(),2);
  weight = TMath::Power(kPi*fXFermionPtr->GetMass()*fXFermionPtr->GetWidth(),2);
#ifdef __NODECAY__
  weight = 1.;
#endif
  bsWeight *= weight;
#endif

  // --------------------------------------------
  //  Calcuate differential cross section
  // --------------------------------------------

  GENBranch x1branch(fQ2X1, fCosThetaEm, fPhiEm, kM_e*kM_e, fMassDM*fMassDM);
  GENBranch x2branch(fQ2X2, fCosThetaEp, fPhiEp, kM_e*kM_e, fMassDM*fMassDM);
  GENBranch cmbranch(fQ2XX, fCosTheta  , fPhi  , &x1branch, &x2branch);

  Double_t sigma = DSigmaDX(cmbranch);

  // --------------------------------------------
  //  Fill plots
  // --------------------------------------------

  Xh_fill( 1, fEcmIP            , (bsWeight*sigma));
  Xh_fill( 2, fCosTheta         , (bsWeight*sigma));
  Xh_fill( 3, fPhi              , (bsWeight*sigma));
  Xh_fill( 4, TMath::Sqrt(fQ2X1), (bsWeight*sigma));
  Xh_fill( 5, fCosThetaEm       , (bsWeight*sigma));
  Xh_fill( 6, fPhiEm            , (bsWeight*sigma));
  Xh_fill( 7, TMath::Sqrt(fQ2X2), (bsWeight*sigma));
  Xh_fill( 8, fCosThetaEp       , (bsWeight*sigma));
  Xh_fill( 9, fPhiEp            , (bsWeight*sigma));
  Xh_fill(10, (Double_t)fJCombI , (bsWeight*sigma));
  Xh_fill(11, (Double_t)fJCombF , (bsWeight*sigma));

  return (bsWeight * sigma);
}

//_____________________________________________________________________________
// --------------------------
//  DSigmaDX
// --------------------------
Double_t EHEHBases::DSigmaDX(GENBranch &cmbranch)
{
  // --------------------------------------------
  //  Phase space
  // --------------------------------------------
  Double_t rs = TMath::Sqrt(cmbranch.GetQ2());
  ANL4DVector qcm(rs, 0., 0., 0.);
  GENFrame    cmframe;

  // CM -> X X
  Double_t q2x1  = cmbranch.GetM12();
  Double_t q2x2  = cmbranch.GetM22();
  Double_t cosx  = cmbranch.GetCosTheta();
  Double_t phix  = cmbranch.GetPhi     ();
  GENPhase2 phaseCM(qcm, q2x1, q2x2, cmframe, cosx, phix, 0);
  ANL4DVector px1 = phaseCM.GetFourMomentum(0);
  ANL4DVector px2 = phaseCM.GetFourMomentum(1);
  Double_t betax  = phaseCM.GetBetaBar();
  if (betax <= 0.) return 0.;

  // X1 -> e- DM1
  GENBranch &x1branch = * cmbranch.GetBranchPtr(0);
  Double_t cosh1 = x1branch.GetCosTheta();
  Double_t phih1 = x1branch.GetPhi     ();
  Double_t mh21  = x1branch.GetM12();
  Double_t mdm21 = x1branch.GetM22();
  GENPhase2 phaseX1(px1, mh21, mdm21, cmframe, cosh1, phih1, 1);
  fP[0]          = phaseX1.GetFourMomentum(0);
  fM[0]          = TMath::Sqrt(mh21);
  fP[1]          = phaseX1.GetFourMomentum(1);
  fM[1]          = TMath::Sqrt(mdm21);
  Double_t betah1 = phaseX1.GetBetaBar();
  if (betah1 <= 0.) return 0.;

  // X2 -> e+ DM2
  GENBranch &x2branch = * cmbranch.GetBranchPtr(1);
  Double_t cosh2 = x2branch.GetCosTheta();
  Double_t phih2 = x2branch.GetPhi     ();
  Double_t mh22  = x2branch.GetM12();
  Double_t mdm22 = x2branch.GetM22();
  GENPhase2 phaseX2(px2, mh22, mdm22, cmframe, cosh2, phih2, 1);
  fP[2]          = phaseX2.GetFourMomentum(0);
  fM[2]          = TMath::Sqrt(mh22);
  fP[3]          = phaseX2.GetFourMomentum(1);
  fM[3]          = TMath::Sqrt(mdm22);
  Double_t betah2 = phaseX2.GetBetaBar();
  if (betah2 <= 0.) return 0.;

  Double_t eb     = rs/2.;
  Double_t pb     = TMath::Sqrt((eb-kM_e)*(eb+kM_e));
  Double_t beta_e = pb/eb;
  fK[0].SetXYZT(0., 0., pb, eb);
  fK[1].SetXYZT(0., 0.,-pb, eb);
#ifdef __DEBUG__
  cerr << "---------------------" << endl;
  cerr << " fP[0] = (" 
       << fP[0].E () << ","
       << fP[0].Px() << ","
       << fP[0].Py() << ","
       << fP[0].Pz() << ")" << endl;
  cerr << " fP[1] = (" 
       << fP[1].E () << ","
       << fP[1].Px() << ","
       << fP[1].Py() << ","
       << fP[1].Pz() << ")" << endl;
  cerr << " fP[2] = (" 
       << fP[2].E () << ","
       << fP[2].Px() << ","
       << fP[2].Py() << ","
       << fP[2].Pz() << ")" << endl;
  cerr << " fP[3] = (" 
       << fP[3].E () << ","
       << fP[3].Px() << ","
       << fP[3].Py() << ","
       << fP[3].Pz() << ")" << endl;
  ANL4DVector qx1 = fP[0] + fP[1];
  ANL4DVector qx2 = fP[2] + fP[3];
  cerr << " ---- " << endl;
  cerr << " rs = " << rs << endl;
  cerr << " mh1 = " << fP[0].GetMass() << endl;
  cerr << " md1 = " << fP[1].GetMass() << endl;
  cerr << " mh2 = " << fP[3].GetMass() << endl;
  cerr << " md2 = " << fP[4].GetMass() << endl;
  cerr << " mx1 = " << qx1.GetMass() << endl;
  cerr << " mx2 = " << qx2.GetMass() << endl;

  ANL4DVector pcm = qx1 + qx2;
  cerr << " pcm = (" 
       << pcm.E () << ","
       << pcm.Px() << ","
       << qcm.Py() << ","
       << qcm.Pz() << ")" << endl;
#endif

  // --------------------------------------------
  //  Calcuate differential cross section
  // --------------------------------------------
  Double_t s      = rs * rs;

  // -------------------
  //  Amplitude squared
  // -------------------
  Double_t amp2 = AmpSquared(cmbranch);

  // -------------------
  //  Put them together
  // -------------------
  static const Int_t    kNbrMx = 3;
#ifndef __NODECAY__
  static const Int_t    kNbr   = kNbrMx;
#else
  static const Int_t    kNbr   = 1;
  betah1 = 1.;
  betah2 = 1.;
#endif
  static const Double_t kFact = k2Pi/(TMath::Power(k4Pi,3*kNbr+kNbrMx-kNbr));

  Double_t identp = 1.;                                // identical particle factor
  Double_t dPhase = kFact * betax * betah1 * betah2;   // phase space factor
  Double_t flux   = 1./(2.* s * beta_e);               // beam flux factor
  Double_t spin   = 1./2.;                             // spin average for e+

  Double_t sigma  = identp * flux * spin * amp2 * dPhase; // in [1/GeV^2]
           sigma *= kGeV2fb;                              // now in [fb]

  return sigma;
}

//_____________________________________________________________________________
// --------------------------
//  AmpSquared()
// --------------------------
Double_t EHEHBases::AmpSquared(GENBranch &cmbranch)
{
  Complex_t amp   = FullAmplitude();
  Double_t  amp2  = TMath::Power(abs(amp),2);

  return amp2;
}

//_____________________________________________________________________________
// --------------------------
//  FullAmplitude()
// --------------------------
Complex_t EHEHBases::FullAmplitude()
{
   Double_t mx     = fXFermionPtr ->GetMass();
   Double_t gamx   = fXFermionPtr ->GetWidth();
   Double_t mdm    = fDMBosonPtr  ->GetMass();
   //-------------------
   // Coupling consts.
   //-------------------
   Double_t Ch     = fZHBosonPtr->GetCh();
   Double_t Sh     = fZHBosonPtr->GetSh();
   //Double_t Cf     = fZHBosonPtr->GetCf();
   //Double_t Sf     = fZHBosonPtr->GetSf();
   //Double_t gxhdm  = -kGw*kGw*Sf*Cf*fF/(kCos2W*2.*TMath::Sqrt(2.))
   //	                  *(Ch*kCosW+Sh*kSinW)*(Sh*kCosW-Ch*kSinW);
#if 0
   cerr << " Ch = " << Ch << " Sh = " << Sh << " Cf = " << Cf << " Sf = " << Sf << endl;
   cerr << " gxhdm = " << gxhdm << " kGw = " << kGw << endl;
#endif 
   //Double_t  qe    = -1.;
   //Double_t  ge    = -qe*kGe;
   //Double_t  glae  = ge;
   //Double_t  grae  = ge;

   //Double_t  t3e   = -1./2.;
   //Double_t  glze  = -kGz*(t3e - qe*kSin2W);
   //Double_t  grze  = -kGz*(    - qe*kSin2W);

   //Double_t  gamz  = fZBosonPtr->GetWidth();

   //Double_t  glehzh = kGw*(Ch-Sh*kSinW/(5.*kCosW))/2.;
   //Double_t  grehzh = 0.;
   Double_t  glehah = kGw*(Sh+Ch*kSinW/(5.*kCosW))/2.;
   Double_t  grehah = 0.;

   static const Bool_t kIsIncoming = kTRUE;
   static const Bool_t kIsOutgoing = kFALSE;

   HELFermion em(fK[0], kM_e, fHelInitial[0], +1, kIsIncoming);
   HELFermion ep(fK[1], kM_e, fHelInitial[1], -1, kIsOutgoing);

#ifndef __NODECAY__
   HELFermion emf(fP[0], kM_e, fHelFinal[0], +1, kIsOutgoing);  // e- from eH-
   HELVector  dm1(fP[1], mdm , fHelFinal[1], +1);               // AH1 from eH-
   HELFermion eHm(emf, dm1, glehah, grehah, mx, gamx);          // eH-

   HELFermion epf(fP[2], kM_e, fHelFinal[2], -1, kIsIncoming);  // e+ from eH+
   HELVector  dm2(fP[3], mdm , fHelFinal[3], +1);               // AH2 from eH+
   HELFermion eHp(epf, dm2, glehah, grehah, mx, gamx);          // eH+
#else
   ANL4DVector px1 = fP[0] + fP[1];
   ANL4DVector px2 = fP[2] + fP[3];
   HELFermion  eHm(px1 , mx , fHelFinal[0], +1, kIsOutgoing);
   HELFermion  eHp(px2 , mx , fHelFinal[2], -1, kIsIncoming);
#endif
   Complex_t amp = AmpEEtoXX(em, ep, eHm, eHp);

   return amp;
}

//_____________________________________________________________________________
// --------------------------
//  AmpEEtoXX()
// --------------------------
Complex_t EHEHBases::AmpEEtoXX(const HELFermion &em,
                               const HELFermion &ep,
                               const HELFermion &eHm,
                               const HELFermion &eHp)
{
   //-------------------
   // Coupling consts.
   //-------------------
   Double_t  Ch     = fZHBosonPtr->GetCh();
   Double_t  Sh     = fZHBosonPtr->GetSh();
   Double_t  qe     = -1.;
   Double_t  ge     = -qe*kGe;
   Double_t  glehzh = kGw*(Ch-Sh*kSinW/(5.*kCosW))/2.;
   Double_t  grehzh = 0.;
   Double_t  glae   = ge;
   Double_t  grae   = ge;

   Double_t  qeH    = -1.;
   Double_t  glaeH  = -qeH*kGe;   //cal
   Double_t  graeH  = -qeH*kGe ;   //cal

   Double_t  t3e   = -1./2.;
   Double_t  glze  = -kGz*(t3e - qe*kSin2W);
   Double_t  grze  = -kGz*(    - qe*kSin2W);

   Double_t  gamz  = fZBosonPtr->GetWidth();

   Double_t  t3eH  = -1./2.;
   Double_t  glzeH = -kGz*(t3eH - qeH*kSin2W); //cal!!
   Double_t  grzeH = glzeH;      //cal!! no diff left right

   Double_t  glehah = kGw*(Sh+Ch*kSinW/(5.*kCosW))/2.;
   Double_t  grehah = 0.;

   //--------------------------------------------------
   // Calculate Amplitudes
   //--------------------------------------------------
#ifdef __PHASESPACE__
   //---------------------------
   // Just Phase Space
   //---------------------------
   Complex_t amp = 1;
#else
   //---------------------------
   // XX Production Amplitude
   //---------------------------
   //--
   // T-channel eH
   //--
   Double_t mzh   = fZHBosonPtr->GetMass();  
   Double_t gamzh = fZHBosonPtr->GetWidth();  
   Double_t mah   = fDMBosonPtr->GetMass();  
   Double_t gamah = fDMBosonPtr->GetWidth();  
#if 0
   cerr << " Ch = " << Ch << " Sh = " << Sh << " Cf = " << Cf << " Sf = " << Sf << endl;
   cerr << " gxhdm = " << gxhdm << " kGw = " << kGw << endl;
#endif 

   HELVector  zht(em, eHm, glehzh, grehzh, mzh, gamzh);  //t-channel ZH
   HELVertex  amptee1(eHp, ep, zht, glehzh, grehzh);

   HELVector  aht(em, eHm, glehah, grehah, mah, gamah);  //t-channel AH
   HELVertex  amptee2(eHp, ep, aht, glehah, grehah);

   //--
   // S-channel Z
   //--
   HELVector zs(em, ep, glze, grze, kM_z, gamz); // s-channel   Z: backward-going
   //   HELVertex ampzxx(eHm, eHp, zs, ?glzeH, ?grzeH);     //--needs to checked---
   HELVertex ampzxx(eHp, eHm, zs, glzeH, grzeH);

   //--
   // S-channel A
   //--
   HELVector as(em, ep, glae, grae,   0.,   0.); // s-channel gamma
   //   HELVertex ampaxx(xm, xp, as, ?glaeH, ?graeH);   //----needs to be checked-----
   HELVertex ampaxx(eHp, eHm, as, glaeH, graeH);

   //--
   // Sum up all the amplitudes
   //--
#if 1
   Complex_t amp = amptee1 + amptee2 - ampzxx - ampaxx;
#else
   Complex_t amp = ampzxx + ampaxx;
#endif
#endif /* end __PHASESPACE__ */

   return amp;
}



//_____________________________________________________________________________
// --------------------------
//  Userin
// --------------------------
void EHEHBases::Userin()
{
  // --------------------------------------------
  //  Open beamstrahlung data
  // --------------------------------------------
  string bsfiledir;
  string bsfilename;

  if (fBeamStr) {
    stringstream ins(gJSF->Env()->GetValue("EHEHBases.BeamstrahlungFilepath",
                                           "/proj/soft/jsf/pro/data/bsdata/"));
    ins >> bsfiledir;

    ins.clear();
    ins.str(gJSF->Env()->GetValue("EHEHBases.BeamstrahlungFilename","trc500"));
    ins >> bsfilename;
    
    TFile *fBeamFile = new TFile((bsfiledir+bsfilename+".root").data());
    if (!fBeamFile) {
      cerr << " Unable to open a file for beamstrahlung" << endl;
      abort();
    }
    
  // --------------------------------------------
  //  Initialize beam generator
  // --------------------------------------------
    JSFBeamGenerationCain::EIBType bstype=JSFBeamGenerationCain::kUniform;
    if ( fIBType == 2 ) { bstype=JSFBeamGenerationCain::kGauss; }

    fBM = (JSFBeamGenerationCain*)fBeamFile->Get(bsfilename.data());
    fBM->SetIBParameters(fBeamWidth, bstype);

    fBM->MakeBSMap();
    fBM->Print(); 

    cout << " Nominal Energy (beamstrahlung) = " << fBM->GetNominalEnergy() 
         << " beam width     (beamstrahlung) ="  << fBM->GetIBWidth() 
         << endl;

    if (!(fBM->GetNominalEnergy() == (fEcmInit/2))) {
      cout << "Nominal energy from beamstrahung is " << fBM->GetNominalEnergy() 
           << " which is different from fEcm/2: "    << (fEcmInit/2)
           << endl;        
    } // check the energy homogeneity
  } // if beamstrahlung is on

  // --------------------------------------------
  //  Initialize Z decay table
  // --------------------------------------------
  if (!fWBosonPtr) fWBosonPtr = new GENPDTWBoson();
  fWBosonPtr->DebugPrint();
  if (!fZBosonPtr) fZBosonPtr = new GENPDTZBoson();
  fZBosonPtr->DebugPrint();
  if (!fPhotonPtr) fPhotonPtr = new GENPDTPhoton();
  //fPhotonPtr->DebugPrint();
  if (!fGluonPtr)  fGluonPtr  = new GENPDTGluon();
  //fGluonPtr->DebugPrint();

  cerr << " WH Boson "      << endl;
  if (!fWHBosonPtr)  fWHBosonPtr  = new WHBoson(fF);
  fWHBosonPtr->DebugPrint();

  cerr << " ZH Boson "      << endl;
  if (!fZHBosonPtr)  fZHBosonPtr  = new ZHBoson(fF, fMassHiggs);
  fZHBosonPtr->DebugPrint();
  fMassDM = fZHBosonPtr->GetMassDM();

  cerr << " AH Boson " << endl;
  if (!fDMBosonPtr) fDMBosonPtr = new AHBoson(fMassDM);
  fDMBosonPtr->DebugPrint();

  cerr << " eH Fermion " << endl;
  if (!fXFermionPtr) fXFermionPtr = new EHFermion(fF,fKappaL);
  fMass   = fXFermionPtr->GetMass();
  fXFermionPtr->DebugPrint();
  Double_t mx = fMass;

  // --------------------------------------------
  //  Define some plots
  // --------------------------------------------
  Xh_init( 1,     0., fEcmInit*1.1, 50, "Ecm"    );
  Xh_init( 2, fXL[0], fXU[0],       50, "Costh"  );
  Xh_init( 3, fXL[1], fXU[1],       50, "Phi"    );
  Xh_init( 4, mx-10., mx+10.,       50, "Mx1"    );
  Xh_init( 5, fXL[2], fXU[2],       50, "CosthH1");
  Xh_init( 6, fXL[3], fXU[3],       50, "PhiEm"  );
  Xh_init( 7, mx-10., mx+10.,       50, "Mx2"    );
  Xh_init( 8, fXL[4], fXU[4],       50, "CosthH2");
  Xh_init( 9, fXL[5], fXU[5],       50, "PhiEp"  );
  Xh_init(10,     0.,     2.,        2, "Helin " );
  Xh_init(11,     0.,    36.,       36, "Helot " );
}

//_____________________________________________________________________________
// --------------------------
//  Userout
// --------------------------
void EHEHBases::Userout()
{
  cout << "End of EHEHBases----------------------------------- "  << endl
       << "Ecm                  = " << fEcmInit << " [GeV]   "    << endl
       << "Beamstrahlung        = " << (fBeamStr ? "on" : "off")  << endl
       << "Bremsstrahlung       = " << (fISR     ? "on" : "off")  << endl
       << "Total Cross section  = " << GetEstimate()  << " +/- "
                                    << GetError()     << " [fb]"  << endl
       << "Number of iterations = " << GetNoOfIterate()           << endl;
}

//_____________________________________________________________________________
// --------------------------
//  SelectHelicities
// --------------------------
void EHEHBases::SelectHelicities(Double_t &weight)
{
   static const Int_t kNi = 2;
   static const Int_t kIHelComb[kNi][2] = {{-1, +1},
                                           {+1, -1}};
   static const Int_t kNf = 36;
   static const Int_t kFHelComb[kNf][4] = {{-1, -1, -1, -1},
                                           {-1, -1, -1,  0},
                                           {-1, -1, -1, +1},
                                           {-1, -1, +1, -1},
                                           {-1, -1, +1,  0},
                                           {-1, -1, +1, +1},

					   {-1,  0, -1, -1},
                                           {-1,  0, -1,  0},
                                           {-1,  0, -1, +1},
                                           {-1,  0, +1, -1},
                                           {-1,  0, +1,  0},
                                           {-1,  0, +1, +1},

					   {-1, +1, -1, -1},
                                           {-1, +1, -1,  0},
                                           {-1, +1, -1, +1},
                                           {-1, +1, +1, -1},
                                           {-1, +1, +1,  0},
                                           {-1, +1, +1, +1},

					   {+1, -1, -1, -1},
                                           {+1, -1, -1,  0},
                                           {+1, -1, -1, +1},
                                           {+1, -1, +1, -1},
                                           {+1, -1, +1,  0},
                                           {+1, -1, +1, +1},

					   {+1,  0, -1, -1},
                                           {+1,  0, -1,  0},
                                           {+1,  0, -1, +1},
                                           {+1,  0, +1, -1},
                                           {+1,  0, +1,  0},
                                           {+1,  0, +1, +1},

					   {+1, +1, -1, -1},
                                           {+1, +1, -1,  0},
                                           {+1, +1, -1, +1},
                                           {+1, +1, +1, -1},
                                           {+1, +1, +1,  0},
                                           {+1, +1, +1, +1}};

   Double_t helm = (1. - fPole)/2.;
   if (fHelCombInitial < helm) {  //fHelCombInitial is random No. between 0 to 1
      fJCombI = 0;
   } else {
      fJCombI = 1;
   }
   fHelInitial[0] = kIHelComb[fJCombI][0];
   fHelInitial[1] = kIHelComb[fJCombI][1];
   fJCombF = (Int_t)(fHelCombFinal*kNf);
   fJCombF = TMath::Min(fJCombF, kNf-1);
   fHelFinal  [0] = kFHelComb[fJCombF][0];
   fHelFinal  [1] = kFHelComb[fJCombF][1];
   fHelFinal  [2] = kFHelComb[fJCombF][2];
   fHelFinal  [3] = kFHelComb[fJCombF][3];
#ifndef __NODECAY__  
   weight = kNf;
#else 
   weight = 4;
#endif
}

//-----------------------------------------------------------------------------
// ==============================
//  class ZHBoson
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
ZHBoson::ZHBoson(Double_t f,
                 Double_t mh)
	: fF(f), fMassHiggs(mh)
{
   fCf          = 1.-4.*kCos2W*kSin2W*kM_z*kM_z/TMath::Power(kGe*f,2);
   fSf          = TMath::Sqrt((1.-fCf)*(1.+fCf));
   Double_t g2  = kGw;
   Double_t gp  = kGe/kCosW;
   Double_t maa = g2*g2*f*f*(fCf*fCf+7.)/8.;
   Double_t mab = g2*gp*f*f*(1.-fCf)*(1.+fCf)/8.;
   Double_t mbb = gp*gp*f*f*(5.*fCf*fCf+3.)/40.;

   fName    = TString("ZH");
   fPID     = 200000003;
   fCharge  =  0.0;
   fSpin    =  1.0;
   fMass    = TMath::Sqrt(0.5*(maa+mbb+TMath::Sqrt(TMath::Power(maa-mbb,2)+4*mab*mab))); 
   fGen     =    0;
   fIsoSpin =  0.0;
   fColor   =  1.0;

   fSh      = -2.*mab/TMath::Sqrt(
                  TMath::Power(maa-mbb+TMath::Sqrt(TMath::Power(maa-mbb,2)+4*mab*mab),2)
                  +4.*mab*mab);
   fCh      = TMath::Sqrt((1.-fSh)*(1.+fSh));
   fMassDM  = TMath::Sqrt(0.5*(maa+mbb-TMath::Sqrt(TMath::Power(maa-mbb,2)+4.*mab*mab)));

   Initialize();
}

//_____________________________________________________________________________
// --------------------------
//  Initialize
// --------------------------
void ZHBoson::Initialize()
{
   //--
   // Couplings
   //--
   Double_t  a  = -kGw*kGw*fSf*fCf*fF/(kCos2W*2.*TMath::Sqrt(2.))
	                  *(fCh*kCosW+fSh*kSinW)*(fSh*kCosW-fCh*kSinW);
   //--
   // ZH --> H + AH
   //--
   GENDecayMode *dmp;
   GENPDTEntry  *d1p, *d2p;
   Int_t    idh   = 25;
   Double_t chgh  = 0.0;
   Double_t spinh = 0.0;
   d1p = new GENPDTEntry("H",idh,chgh,spinh,fMassHiggs);
   d2p = new AHBoson(fMassDM);
   Double_t m1    = d1p->GetMass();
   Double_t m2    = d2p->GetMass();
   Double_t ident = 1.;
   Double_t gam = GamToSV(m1, m2, a)/ident;
   if (gam > 0.) {
      dmp = new GENDecayMode(gam);
      dmp->Add(d1p);
      dmp->Add(d2p);
      Add(dmp);
   }
}

//_____________________________________________________________________________
// --------------------------
//  GamToSV
// --------------------------
Double_t ZHBoson::GamToSV(Double_t m1, // 1st daughter mass
                          Double_t m2, // 2nd daughter mass
                          Double_t a)  // coupling
{
   Double_t x1   = TMath::Power(m1/fMass,2);
   Double_t x2   = TMath::Power(m2/fMass,2);
   Double_t beta = 1. - 2.*(x1+x2) + TMath::Power((x1-x2),2);

   if (beta <= 0.) return 0.;
   beta = TMath::Sqrt(beta);

   ANL4DVector px(fMass,0.,0.,0.);
   double p = fMass*beta/2;
   ANL4DVector ph (sqrt(p*p+m1*m1),0.,0., p);
   ANL4DVector pdm(sqrt(p*p+m2*m2),0.,0.,-p);
   HELScalar h(ph, +1);
   HELVector *xPtr[3], *dPtr[3];
   for (Int_t i=0; i<3; i++) {
     xPtr[i] = new HELVector(px , fMass, i-1, -1);
     dPtr[i] = new HELVector(pdm, m2   , i-1, +1);
   } 
   Double_t amp2 = 0.;
   for (Int_t jx=0; jx<3; jx++) {
      for (Int_t jd=0; jd<3; jd++) {
         HELVertex amp(*xPtr[jx], *dPtr[jd], h, a);
         amp2 += pow(abs(amp),2);
      }
   }
   Double_t spin = 2*fSpin+1;
   Double_t fac  = 1./(16.*kPi)/fMass/spin;
   Double_t gam  = fac*amp2*beta;

   return gam;
}


//-----------------------------------------------------------------------------
// ==============================
//  class WHBoson
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
WHBoson::WHBoson(Double_t f)
	: fF(f)
{
   Double_t Cf  = 1.-4.*kCos2W*kSin2W*kM_z*kM_z/TMath::Power(kGe*f,2);
   Double_t g2  = kGw;
   Double_t gp  = kGe/kCosW;
   Double_t maa = g2*g2*f*f*(Cf*Cf+7.)/8.;
   Double_t mab = g2*gp*f*f*(1.-Cf)*(1.+Cf)/8.;
   Double_t mbb = gp*gp*f*f*(5.*Cf*Cf+3.)/40.;

   fName    = TString("WH");
   fPID     = 200000000;
   fCharge  = -1.0;
   fSpin    =  1.0;
   fMass    =  g2*f*TMath::Sqrt(Cf + 3.)/2.;
   fGen     =    0;
   fIsoSpin =  1.0;
   fColor   =  1.0;

   fSh      = -2.*mab/TMath::Sqrt(
                  TMath::Power(maa-mbb+TMath::Sqrt(TMath::Power(maa-mbb,2)+4*mab*mab),2)
                  +4.*mab*mab);
   fMassDM  = TMath::Sqrt(0.5*(maa+mbb-TMath::Sqrt(TMath::Power(maa-mbb,2)+4.*mab*mab)));

   Initialize();
}

//_____________________________________________________________________________
// --------------------------
//  Initialize
// --------------------------
void WHBoson::Initialize()
{
   //--
   // Coupling
   //--
   Double_t  a  = kGw*fSh;
   //--
   // WH --> AH + W
   //--
   GENDecayMode *dmp;
   GENPDTEntry  *d1p, *d2p;
   d1p = new AHBoson(fMassDM);
   d2p = new GENPDTWBoson();
   Double_t m1    = d1p->GetMass();
   Double_t m2    = d2p->GetMass();
   Double_t ident = 1.;
   Double_t gam = GamToVV(m1, m2, a)/ident;
   if (gam > 0.) {
      dmp = new GENDecayMode(gam);
      dmp->Add(d1p);
      dmp->Add(d2p);
      Add(dmp);
   }
}

//_____________________________________________________________________________
// --------------------------
//  GamToVV
// --------------------------
Double_t WHBoson::GamToVV(Double_t m1, // 1st daughter mass
                          Double_t m2, // 2nd daughter mass
                          Double_t a)  // coupling
{
   Double_t x1   = TMath::Power(m1/fMass,2);
   Double_t x2   = TMath::Power(m2/fMass,2);
   Double_t beta = 1. - 2.*(x1+x2) + TMath::Power((x1-x2),2);

   if (beta <= 0.) return 0.;
   beta = TMath::Sqrt(beta);

   ANL4DVector px(fMass,0.,0.,0.);
   double p = fMass*beta/2;
   ANL4DVector pw(sqrt(p*p+m2*m2),0.,0.,p);
   ANL4DVector pdm(sqrt(p*p+m1*m1),0.,0.,-p);
   HELVector *xPtr[3], *dPtr[3], *wPtr[3];
   for (Int_t i=0; i<3; i++) {
     xPtr[i] = new HELVector(px , fMass, i-1, -1);
     dPtr[i] = new HELVector(pdm, m1   , i-1, +1);
     wPtr[i] = new HELVector(pw , m2   , i-1, +1);
   } 
   Double_t amp2 = 0.;
   for (Int_t jx=0; jx<3; jx++) {
      for (Int_t jd=0; jd<3; jd++) {
         for (Int_t jw=0; jw<3; jw++) {
            HELVertex amp(*wPtr[jw], *xPtr[jx], *dPtr[jd], a);
            amp2 += pow(abs(amp),2);
	 }
      }
   }
   Double_t spin = 2*fSpin+1;
   Double_t fac  = 1./(16.*kPi)/fMass/spin;
   Double_t gam  = fac*amp2*beta;

   return gam;
}


//-----------------------------------------------------------------------------
// ==============================
//  class AHBoson
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
AHBoson::AHBoson(Double_t m)
{
   fName    = TString("AH");
#if 0
   fPID     = 200000001;
#else
   fPID     = 220000; // LSP code for JSFHadronizer.
#endif
   fCharge  =  0.0;
   fSpin    =  1.0;
   fMass    =    m;
   fGen     =    0;
   fIsoSpin =  0.0;
   fColor   =  1.0;
}

//-----------------------------------------------------------------------------
// ==============================
//  class EHFermion
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
EHFermion::EHFermion(Double_t F, Double_t kappa, Double_t mh)
  : fF(F), fKappa(kappa), fMassHiggs(mh)   //->changed header file
{
   fName    = TString("eH");
   fPID     = 100000011;
   fCharge  = -1.0;
   fSpin    =  0.5;
   fMass    = TMath::Sqrt(2)*fF*fKappa;   //CHECK !!!->checked
   fGen     =    1;   //generation number
   fIsoSpin =   -0.5;
   fColor   =   1.;

   Double_t fCf = 1.-4.*kCos2W*kSin2W*kM_z*kM_z/TMath::Power(kGe*fF,2);
   //Double_t fSf = TMath::Sqrt((1.-fCf)*(1.+fCf));
   Double_t g2  = kGw;
   Double_t gp  = kGe/kCosW;
   Double_t maa = g2*g2*fF*fF*(fCf*fCf+7.)/8.;
   Double_t mab = g2*gp*fF*fF*(1.-fCf)*(1.+fCf)/8.;
   Double_t mbb = gp*gp*fF*fF*(5.*fCf*fCf+3.)/40.;
   fSh      = -2.*mab/TMath::Sqrt(
                  TMath::Power(maa-mbb+TMath::Sqrt(TMath::Power(maa-mbb,2)+4*mab*mab),2)
                  +4.*mab*mab);
   fCh      = TMath::Sqrt((1.-fSh)*(1.+fSh));
   fMassDM  = TMath::Sqrt(0.5*(maa+mbb-TMath::Sqrt(TMath::Power(maa-mbb,2)+4.*mab*mab)));

   Initialize();
}


void EHFermion::Initialize()
{
   //--
   // Coupling
   //--
   Double_t  glehzh = kGw*(fCh-fSh*kSinW/(5.*kCosW))/2.;
   Double_t  grehzh = 0.;
   Double_t  glehah = kGw*(fSh+fCh*kSinW/(5.*kCosW))/2.;
   Double_t  grehah = 0.;
   Double_t  glw    = -kGw*kSqh;
   Double_t  grw    = 0.;

   //--
   // eH --> AH + e
   //--
   GENDecayMode *dmp;
   GENPDTEntry  *d1p, *d2p;
   Int_t    pide   = 11;
   Double_t chge   = -1.;
   Double_t spine  = 1./2.;
   Double_t masse  = kM_e;
   Int_t    gene   = 1;
   Double_t ispine = -1./2.;
   Double_t colore = 1.;
   d1p = new AHBoson(fMassDM);
   d2p = new GENPDTEntry("e",pide,chge,spine,masse,gene,ispine,colore);
   Double_t m1    = d1p->GetMass();
   Double_t m2    = d2p->GetMass();
   Double_t gam = GamToFV(m1, m2, glehah, grehah);
   if (gam > 0.) {
      dmp = new GENDecayMode(gam);
      dmp->Add(d1p);  //add AH to dmp array (decay mode pointer)
      dmp->Add(d2p);  //add e to dmp array (decay mode pointer)
      Add(dmp);   //add dmy to array EHFermion ->mode picker 
   }
#if 1
   //--
   // eH --> ZH + e
   //--
   d1p = new ZHBoson(fF,fMassHiggs);
   d2p = new GENPDTEntry("e",pide,chge,spine,masse,gene,ispine,colore);
   m1  = d1p->GetMass();
   m2  = d2p->GetMass();
   gam = GamToFV(m1, m2, glehzh, grehzh);
   if (gam > 0.) {
      dmp = new GENDecayMode(gam);
      dmp->Add(d1p);  //add AH to dmp array (decay mode pointer)
      dmp->Add(d2p);  //add e to dmp array (decay mode pointer)
      Add(dmp);   //add dmy to array EHFermion ->mode picker 
   }
   //--
   // eH --> WH + nu_e
   //--
   Int_t    pidn   = 1;
   Double_t chgn   = 0.;
   Double_t spinn  = 1./2.;
   Double_t massn  = 0.;
   Int_t    genn   = 1;
   Double_t ispinn = +1./2.;
   Double_t colorn = 1.;
   d1p = new WHBoson(fF);
   d2p = new GENPDTEntry("nu_e",pidn,chgn,spinn,massn,genn,ispinn,colorn);
   m1  = d1p->GetMass();
   m2  = d2p->GetMass();
   gam = GamToFV(m1, m2, glw, grw);
   if (gam > 0.) {
      dmp = new GENDecayMode(gam);
      dmp->Add(d1p);  //add AH to dmp array (decay mode pointer)
      dmp->Add(d2p);  //add e to dmp array (decay mode pointer)
      Add(dmp);   //add dmy to array EHFermion ->mode picker 
   }
#endif
}


//_____________________________________________________________________________
// --------------------------
//  GamToFV
// --------------------------
Double_t EHFermion::GamToFV(Double_t m1, // 1st daughter mass
                          Double_t m2, // 2nd daughter mass
                          Double_t gl,Double_t gr)  // coupling
{
  //Double_t fMass =fXFermionPtr->GetMass(); 
   Double_t x1   = TMath::Power(m1/fMass,2);
   Double_t x2   = TMath::Power(m2/fMass,2);
   Double_t beta = 1. - 2.*(x1+x2) + TMath::Power((x1-x2),2);

   if (beta <= 0.) return 0.;
   beta = TMath::Sqrt(beta);

   ANL4DVector px(fMass,0.,0.,0.);
   double p = fMass*beta/2;
   ANL4DVector pe(sqrt(p*p+m2*m2),0.,0.,p);
   ANL4DVector pdm(sqrt(p*p+m1*m1),0.,0.,-p);
   HELVector  *dPtr[3];
   HELFermion *xPtr[2], *ePtr[2];
   for (Int_t i=0; i<3; i++) {    
     dPtr[i] = new HELVector(pdm, m1, i-1, +1);
   }

   static const Bool_t kIsIncoming =kTRUE;
   static const Bool_t kIsOutgoing =kFALSE;
   xPtr[0] = new HELFermion(px , fMass,+1,+1,kIsIncoming); // eH- Righthanded
   xPtr[1] = new HELFermion(px , fMass,-1,+1,kIsIncoming); // eH- Lefthanded
   ePtr[0] = new HELFermion(pe , kM_e ,+1,+1,kIsOutgoing); // e-  Righthanded
   ePtr[1] = new HELFermion(pe , kM_e ,-1,+1,kIsOutgoing); // e- Lefthanded

   Double_t amp2 = 0.;
   for (Int_t jx=0; jx<2; jx++) {
      for (Int_t jd=0; jd<3; jd++) {
         for (Int_t je=0; je<2; je++) {
            HELVertex amp(*xPtr[jx], *ePtr[je], *dPtr[jd], gl,gr);
            amp2 += pow(abs(amp),2);
	 }
      }
   }
   Double_t spin = 2*fSpin+1;
   Double_t fac  = 1./(16.*kPi)/fMass/spin;
   Double_t gam  = fac*amp2*beta;

   return gam;
}
