//*************************************************************************
//* ----------------
//* U4CDCRawData Class
//* ----------------
//* (Description)
//*	This set of classes reads U4CDC raw data for Lorentz Angle 
//*	measurments and unpacks them into root objects.
//*	When a root file is opened in the read/write mode, the root
//*	objects are stored in it.  If the root file is read-only,
//*	the data are obtained from it.
//*	The package contains the following classes:
//*	The following parameters can be specified in "jsf.conf". 
//*
//*	    U4CDCRawData.RFNFormat:       printf format to construct file name
//*	    U4CDCRawData.DoPlots:         (TRUE/FASLE)
//* (Requires)
//*	TObject:
//*	TClonesArray:
//*	U4CDCUniversalRecordHeader:
//*	U4CDCRawGPIBDatum:
//*	U4CDCRawScalerDatum:
//*	U4CDCRawADCDatum:
//* (Provides)
//*	U4CDCHardwareMap		: hardware map
//*	U4CDCRawDataBuf		: raw data buffer
//*	U4CDCBeginRunBuf		: begin run buffer
//*	U4CDCEndRunBuf		: end run buffer
//*	U4CDCRawData			: steering class
//* (Update Record)
//*	1999/06/02  K.Fujii	Original version.
//*	2000/05/17  K.Fujii	Implementation for new format (not tested)
//*     2001/05/27  K.Fujii     Modification for new branch scheme
//*                             started from ROOT v.3.01.
//*	2001/09/14  K.Fujii     Use automatic schema evolution.
//*				Reverted branch name to U4CDCRawData-EventBuf
//*				to use legacy unpacked data.
//*	2001/11/15  K.Hoshina   Added U4CDCRawGPIBDatum and U4CDCRawScalerDatum.
//*     2002/02/10  K.Fujii     Protect against invalid access to 
//*                             non-existent entries in Hardware Map.
//*                             A bug fix in dtor of U4CDCRawData: if (!fd)
//*                             should have been if (fd).
//*	2002/12/04  K.Fujii	Don't "delete fEventBuf" in ~U4CDCRawData
//*				since it's deleted in ~JSFModule.
//*************************************************************************

#include "JSFSteer.h"
#include "U4CDCRawData.h"
#include "TRandom.h"
#include <iostream>

ClassImp(U4CDCChannel)
ClassImp(U4CDCHardwareMap)
ClassImp(U4CDCBeginRunBuf)
ClassImp(U4CDCRawDataBuf)
ClassImp(U4CDCEndRunBuf)
ClassImp(U4CDCRawData)

#define __DEBUG__

//___________________________________________________________________________
// ----------------
// Begin Run Buffer
// ----------------
//
U4CDCBeginRunBuf::U4CDCBeginRunBuf() : fDateTime(0) {}

//___________________________________________________________________________
U4CDCBeginRunBuf::~U4CDCBeginRunBuf() {}

//___________________________________________________________________________
Bool_t U4CDCBeginRunBuf::ReadFile(FILE *fd) {
    fHeader.ReadFile(fd);
    if ( fHeader.GetRecordType() != kRecTypeBRR ) return kFALSE;
    Len_t   nw   = fHeader.GetDataLength();
    std::cerr << "nw = " << nw << std::endl;
    BRR_t  *data = new BRR_t[nw];
    Len_t   ne   = fread(data, sizeof(BRR_t), nw, fd);
    if ( ne != nw ) {
        Error("BeginRunBuf","Failed to read BRR: nw =%d",ne);
        delete [] data;
        return kFALSE;
    }
    fHeader.DebugPrint();
    Unpack(data);
    delete [] data;
    return kTRUE;
}

//___________________________________________________________________________
void U4CDCBeginRunBuf::Unpack(BRR_t data[])
{
    Int_t fmtid = fHeader.GetFormatID();
    Len_t nw    = fHeader.GetDataLength();
    Int_t ic;
    switch (fmtid) {
        case 0: {				// Old 98 format
            ic = 3;
            break;
        }
        case 1: 				// New format
        case 2: {				// New format
        
            Int_t id = 0;
            UnpackDateTime(&data[id]);
            Int_t im = 2;
            UnpackHardwareMap(&data[im]);
            Int_t it = im + data[im];
            UnpackThresholdData(&data[it]);
            ic = it + data[it];
            break;
        }
        default: { 
            std::cerr << "Format ID = " << fmtid
                 << " is not supported" << std::endl;
            return;
        }
    }
    Char_t *comment =  new Char_t[2*(nw-ic)];
    sscanf((Char_t *)&data[ic],"%s",comment);
    fComment.SetString(comment);
    std::cout << "--------------------" << std::endl
         << " Comment of the Run" << std::endl
         << "--------------------" << std::endl;
    std::cout << fComment.GetString() << std::endl;
}

//___________________________________________________________________________
void U4CDCBeginRunBuf::UnpackDateTime(BRR_t data[])
{
    UShort_t dum[2];
    dum[0] = data[0];
    dum[1] = data[1];
#ifndef __LITTLE_ENDIAN__
    dum[0] = (dum[0] << 8) | (dum[0] >> 8);
    dum[1] = (dum[1] << 8) | (dum[1] >> 8);
    UInt_t dum4 = dum[1];
    fDateTime = (dum4 << 16) | dum[0];
#else
    UInt_t dum4 = dum[0];
    fDateTime = (dum4 << 16) | dum[1]; 
#endif
}

//___________________________________________________________________________
void U4CDCBeginRunBuf::UnpackHardwareMap(BRR_t data[])
{
    Len_t nw;
    Int_t i;
#ifndef __LITTLE_ENDIAN__
    nw = (UShort_t)(data[0] << 8) | (data[0] >> 8);
    for (i=0; i<nw; i++) {
    	data[i] = (data[i] << 8) | (data[i] >> 8);
    }
#else
    nw = data[0];
#endif
    UShort_t dum1, dum2;
    UShort_t nch = nw/2;
    for (i=0; i<nch; i++) {
        dum1 = data[2*i+1];
        dum2 = data[2*i+2];
        Int_t crtno = (UShort_t)(dum1 >> 9);
        Int_t stano = (UShort_t)((UShort_t)(dum1 << 7) >> 11);
        Int_t chano = (UShort_t)((UShort_t)(dum1 << 12) >> 12);
        Int_t id    = (UShort_t)(dum2 >> 14);
        Int_t detch = (UShort_t)((UShort_t)(dum2 << 2) >> 2) - 1;
        U4CDCChannel *chap = new U4CDCChannel(id,detch);
        fMap.Set(crtno,stano,chano) = chap;

#if 1	
	std::cerr << "Hardware Map ID, st#, ch#, dtc# = " 
	     << id << " " << stano << " " << chano << " " << detch << std::endl;
#endif        
        
    }
}

//___________________________________________________________________________
void U4CDCBeginRunBuf::UnpackThresholdData(BRR_t data[])
{
    Int_t i;
#ifndef __LITTLE_ENDIAN__
    Int_t nw = (UShort_t)(data[0] << 8) | (data[0] >> 8);
    for (i=0; i<nw; i++) {
    	data[i] = (data[i] << 8) | (data[i] >> 8);
    }
#else
    Int_t nw = data[0];
#endif
    fThresh.Set(nw-1);
    for (i=0; i<nw-1; i++) {
        fThresh[i] = (UShort_t)data[i+1];
    }
}
//___________________________________________________________________________
void U4CDCBeginRunBuf::EChannelToDChannel(Int_t cr, Int_t st, Int_t ch, 
  			    		Int_t &id, Int_t &dch)
{
    Int_t fmtid = fHeader.GetFormatID();
    if (fmtid) {
        U4CDCChannel *chap = (U4CDCChannel *)fMap(cr,st,ch);
        if (chap) {
            id  = chap->GetID();
            dch = chap->GetDetCh();
        } else {
            std::cerr << ">>>>>> Error in U4CDCBeginRunBuf::EChannelToDChannel..." 
                 << std::endl
                 << " (cr,st,ch) = (" << cr << "," << st << "," << ch << ")"
                 << " is non-existent!" << std::endl
                 << " ----> id and dch set to 9999." << std::endl;
            id  = 99999999;
            dch = 99999999;
        }
    } else {
        if (ch < 12) {
	   id  = 0;
           dch = ch;
        } else if (ch == 14) {
           id  = 1;
           dch = ch;
        } else {
            std::cerr << ">>>>>> Error in U4CDCBeginRunBuf::EChannelToDChannel..." 
                 << std::endl
                 << " ch = " << ch 
                 << " is non-existent in old format!" << std::endl
                 << " ----> id and dch set to 9999." << std::endl;
           id  = 99999999;
           dch = 99999999;
        }
    }
}


//___________________________________________________________________________
// ---------------
// Raw Data Buffer
// ---------------
//
U4CDCRawDataBuf::U4CDCRawDataBuf(const char *name, const char *title, 
	U4CDCRawData *module) : JSFEventBuf(name, title, (JSFModule*)module),
			      fNGPIBData(0), fNScalerData(0), fNADCData(0), 
			      fGPIBData(0), fScalerData(0), fADCData(0),
			      fNonFADCIndex(-999999) {}

//___________________________________________________________________________
U4CDCRawDataBuf::U4CDCRawDataBuf(U4CDCRawData *module, const char *name, 
	const char *title) : JSFEventBuf(name, title, (JSFModule*)module),
			     fNGPIBData(0), fNScalerData(0), fNADCData(0), 
			     fGPIBData(0), fScalerData(0), fADCData(0),
			     fNonFADCIndex(-999999) {}

//___________________________________________________________________________
U4CDCRawDataBuf::~U4CDCRawDataBuf()
{
    if ( fGPIBData )   delete fGPIBData;
    if ( fScalerData ) delete fScalerData;
    if ( fADCData )    delete fADCData;
}

//___________________________________________________________________________
Bool_t U4CDCRawDataBuf::ReadFile(FILE *fd)
{
  U4CDCUniversalRecordHeader header;
  for (Int_t i=0; i<3; i++) {
    if ( !fHeader.ReadFile(fd) ) return kFALSE;
    switch ( fHeader.GetRecordType() ) {
	case kRecTypeBRR:
	case kRecTypeERR:
		return kFALSE;
    }
    Int_t recID = fHeader.GetRecordType(); 
    Len_t   nw  = fHeader.GetDataLength();
    ADC_t  *data = new ADC_t[nw];
    Len_t   ne   = fread(data, sizeof(ADC_t), nw, fd);
    if ( ne != nw ) {
        Error("U4CDCRawDataBuf","Failed to read Event Data: nw =%d",ne);
        delete [] data;
        return kFALSE;
    }
    SetEventNumber(fHeader.GetEventNumber());
    switch ( recID ) {
	case kRecTypeADC:
		header = fHeader;
   		U4CDCRawADCDatum::SetFormat(fHeader);
    		fNADCData = nw/U4CDCRawADCDatum::GetRawLength();
    		Unpack(data);
		break;
	case kRecTypeGPIB:
    		U4CDCRawGPIBDatum::SetFormat(fHeader);
    		fNGPIBData = nw/U4CDCRawGPIBDatum::GetRawLength();
    		Unpack(data);
		break;
	case kRecTypeScaler:
    		U4CDCRawScalerDatum::SetFormat(fHeader);
    		fNScalerData = nw/U4CDCRawScalerDatum::GetRawLength();
    		Unpack(data);
		break;
	default:
		if (recID > 0) {
        		Error("U4CDCRawDataBuf","Record Type not supported: Type=%d",recID);
		}
    		delete [] data;
    		return kFALSE;
    }
    delete [] data;
    //  >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>   Oo. >')))<
    if (fHeader.GetFormatID() != 2) break;
    //  >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>   Oo. >')))<
  }
  fHeader = header;
  return kTRUE;
}

//___________________________________________________________________________
void U4CDCRawDataBuf::Unpack(ADC_t data[])
{
    switch ( fHeader.GetRecordType() ) {
	case kRecTypeGPIB: {
    		if ( !fGPIBData ) fGPIBData = new TClonesArray("U4CDCRawGPIBDatum",fNGPIBData);
    		TClonesArray & gpibdata = *fGPIBData;
    		Len_t lgpib = U4CDCRawGPIBDatum::GetRawLength();
    		for ( Int_t i = 0; i < fNGPIBData; i++ ) {
        		new(gpibdata[i]) U4CDCRawGPIBDatum(&data[lgpib*i]);
    		}
		break;
	}
	case kRecTypeScaler: {
    		if ( !fScalerData ) fScalerData = new TClonesArray("U4CDCRawScalerDatum",fNScalerData);
    		TClonesArray & scalerdata = *fScalerData;
    		Len_t lscaler = U4CDCRawScalerDatum::GetRawLength();
    		for ( Int_t i = 0; i < fNScalerData; i++ ) {
        		new(scalerdata[i]) U4CDCRawScalerDatum(&data[lscaler*i]);
    		}
		break;
	}
	case kRecTypeADC: {
		U4CDCBeginRunBuf *br = ((U4CDCRawData *)fModule)->GetBeginRunBuf();
    		if ( !fADCData ) fADCData = new TClonesArray("U4CDCRawADCDatum",fNADCData);
    		TClonesArray & adcdata = *fADCData;
    		Len_t ladc = U4CDCRawADCDatum::GetRawLength();
		fNonFADCIndex = -999999;
    		for ( Int_t i = 0; i < fNADCData; i++ ) {
        		new(adcdata[i]) U4CDCRawADCDatum(&data[ladc*i]);
			if (fNonFADCIndex < 0) {
				Int_t cr = 0; // there is only 1 crate now
				Int_t st = ((U4CDCRawADCDatum *)adcdata[i])->GetStationNo();
				Int_t ch = ((U4CDCRawADCDatum *)adcdata[i])->GetChannelNo();
				Int_t id = 0, a;
				br->EChannelToDChannel(cr,st,ch,id,a);
				if (id) fNonFADCIndex = i;
			}
    		}
	}
    }
}
//___________________________________________________________________________
#if 0
void U4CDCRawDataBuf::Streamer(TBuffer &R__b)
{
   // Stream an object of class U4CDCRawDataBuf.

   if (R__b.IsReading()) {
      UInt_t R__s, R__c;
      Version_t R__v = R__b.ReadVersion(&R__s, &R__c);
      if (R__v > 1) {
        U4CDCRawDataBuf::Class()->ReadBuffer(R__b, this, R__v, R__s, R__c);
        return;
      }
   } else {
      U4CDCRawDataBuf::Class()->WriteBuffer(R__b, this);
   }
}
#endif

//___________________________________________________________________________
// --------------
// End Run Buffer
// --------------
//
Bool_t U4CDCEndRunBuf::ReadFile(FILE *fd) {
    fHeader.ReadFile(fd);
    if ( fHeader.GetRecordType() != kRecTypeERR ) return kFALSE;
    Len_t  nw   = fHeader.GetDataLength();
    ERR_t *data = new ERR_t[nw];
    Len_t  ne = fread(data, sizeof(ERR_t), nw, fd);
    if ( ne != nw ) {
        Error("BeginRunBuf","Failed to read ERR: nw =%d",ne);
        delete [] data;
        return kFALSE;
    }
    Unpack(data);
    delete [] data;
    return kTRUE;
}
//___________________________________________________________________________
void U4CDCEndRunBuf::Unpack(ERR_t data[])
{
}

//___________________________________________________________________________
// -----------------
// Raw Data Analysis
// -----------------
//
U4CDCRawData::U4CDCRawData(const char *name, const char *title)
       : JSFModule(name,title)
{
    std::cerr << " constructor for U4CDCRawData is called..." << std::endl;
    fBeginBuf = new U4CDCBeginRunBuf();
    fEventBuf = new U4CDCRawDataBuf(this);

    Len_t len = strlen( gJSF->Env()->GetValue("U4CDCRawData.RFNFormat",
                         "/proj/cdc/data/baby/2001/beam/Run%5.5d.dat") );
    fRFNFormat = new Char_t[len];
    strcpy( fRFNFormat,  gJSF->Env()->GetValue("U4CDCRawData.RFNFormat",
                         "/proj/cdc/data/baby/2001/beam/Run%5.5d.dat") );
    fd = NULL;
    SetBufferSize(2000000);  // buffer size for event data.
}

//___________________________________________________________________________
U4CDCRawData::~U4CDCRawData()
{
    std::cerr << " U4CDCRawData will be deleted...fEventBuf is " 
    	 << (Int_t)fEventBuf << std::endl;
    delete fBeginBuf;
    delete fRFNFormat;
    if(fd) fclose(fd);
}

//___________________________________________________________________________
Bool_t U4CDCRawData::BeginRun(Int_t nrun)
{
    std::cout << " BeginRun is called for run " << nrun << std::endl;
        
    // Open Raw Data Root File
   
    if ( ! fFile->IsWritable() ) {
    	TDirectory *last = gDirectory;
    	Char_t dir[100];
    	sprintf(dir,"/conf/begin%5.5i",nrun);
    	fFile->cd(dir);
    	std::cout << " current file name is " << fFile->GetPath() << std::endl;
    	gDirectory->pwd();
    	fBeginBuf->Read("U4CDCBeginRunBuf");
	std::cout << " 	BRR read from it " << std::endl;
    	last->cd();
        return kTRUE;
    }

    // Open Raw Data File
    
    Len_t ln = strlen(fRFNFormat);
    Char_t *infile = new Char_t[ln+10];
    sprintf(infile,fRFNFormat,nrun);
    std::cout << " Going to open a file " << infile << std::endl;

    fd = fopen(infile,"rb");
    std::cerr << " fd = " << (void*)fd << std::endl;

    if ( fd == NULL ) {
        Error("BeginRun","Failed to open %s.",infile);
        delete [] infile;
        return kFALSE;
    }
    delete [] infile;
    
    // Read and Unpack Begin Run Record
    
    std::cout << " Going to unpack BRR " << std::endl;
    if ( ! fBeginBuf->ReadFile(fd) ) return kFALSE;
    
    Int_t irun = fBeginBuf->GetHeader()->GetRunNumber();
    std::cout << " BRR unpacked " << std::endl;
  
    // output standard information
    
    SetModuleStatus(kBeginRun);
    SetRunNumber(irun);
    ConfDirectory();
    
    // Save begin run record
    
    TDirectory *last = gDirectory;
    Char_t dir[100];
    sprintf(dir,"/conf/begin%5.5i",nrun);
    fFile->cd(dir);
    std::cout << " current file name is " << fFile->GetPath() << std::endl;

    gDirectory->pwd();
    fBeginBuf->Write("U4CDCBeginRunBuf");
    last->cd();

    return kTRUE;
}

//___________________________________________________________________________
Bool_t U4CDCRawData::Process(Int_t ev)
{
    U4CDCRawDataBuf *buf = (U4CDCRawDataBuf*)fEventBuf;
    if ( fFile->IsWritable() ) {
REPEAT:
        while (!buf->ReadFile(fd)) {
            Int_t id = buf->GetHeader()->GetRecordType();
            switch (id) {
                case kRecTypeERR: {
                    std::cout << "ERR encountered " << std::endl;
                    SetModuleStatus(kEndRun);
                    return kFALSE;
                }
                case -9999: {
                    std::cout << "EOF encountered " << std::endl;
                    return kFALSE;
                }
                default: {
                    std::cout << " Obtained record type is " << id << std::endl;
                }
            }
        }
    
        if (gJSF->GetRunNumber() != buf->GetHeader()->GetRunNumber()) {
            std::cerr << "????? ERROR ????? Run number inconsistency" << std::endl;
            buf->GetHeader()->DebugPrint();
            goto REPEAT;
        }
    }
    buf->GetHeader()->DebugPrint();

    return kTRUE;
}

//___________________________________________________________________________
void U4CDCRawData::MakeBranch(TTree *tree)
{   
   fTree=tree;
   if( fEventBuf && fMakeBranch ) {
#if ROOT_VERSION_CODE >= ROOT_VERSION(3,1,3)
     Int_t split=-1;
#else
     Int_t split=0;
#endif
     Int_t bsize= 200000;
//
// 2001/09/14  Reverted to old branch name to use legacy unpacked data.
#if 0
     tree->Branch(fEventBuf->GetName(), fEventBuf->ClassName() ,
                  &fEventBuf, bsize, split);
#else
     Char_t name[50];
     sprintf(name,"%s-EventBuf",GetName());
     tree->Branch(name, fEventBuf->ClassName() ,
                  &fEventBuf, bsize, split);
#endif
   }
}

//___________________________________________________________________________
void U4CDCRawData::SetBranch(TTree *tree)
{   
   fTree=tree;
   if( fEventBuf ) {
     Char_t name[50];
//
// 2001/09/14  Reverted to old branch name to use legacy unpacked data.
#if 0
     sprintf(name,"%s",fEventBuf->GetName());
#else
     sprintf(name,"%s-EventBuf",GetName());
#endif
     fBranch=tree->GetBranch(name);
#if 1
     fBranch->SetAddress(&fEventBuf);
#else
     if (fBranch) fBranch->SetAddress(&fEventBuf);
#endif
   }
}

