//*************************************************************************
//* ---------------------
//* U4ITTrack Class
//* ---------------------
//* (Description)
//*	Class to store "U4ITKalHit"s for a single track.
//* (Requires)
//*	TObjArray
//* (Provides)
//*	U4ITTrack		: Array containing "U4ITKalHit"s
//* (Update Record)
//*     2004/08/26   A.Yamaguchi   Original Version. 
//*************************************************************************

#include <iostream>
#include <iomanip>

#include "U4ITTrack.h"
#include "U4ITKalHit.h"
#include "U4ITFBKalHit.h"
#include "U4ITTrackHit.h"
#include "U4ITMeasLayer.h"
#include "U4ITFBMeasLayer.h"
#include "TKalTrackSite.h"
using namespace std;
//#define __DEBUG__

ClassImp(U4ITTrack)

//=====================================================================
//* constructor -------------------------------------------------------

U4ITTrack::U4ITTrack()
{
   SetOwner();
}

U4ITTrack::U4ITTrack(U4ITTrackMaker* trkm)
         : U4VTrack(trkm)
{
   SetOwner();
}

U4ITTrack::U4ITTrack(const U4ITTrack &tk)
         : U4VTrack(tk)
{
   // Caution: 
   //   The derived class should delete the U4ITTrackHit's in the
   //   this and new TrackHits corresponding to the derived class.

   TIter next(&tk);
   U4ITTrackHit *thitp = 0;
   while ((thitp = (U4ITTrackHit *)next())) {
      Add(new U4ITTrackHit(*thitp));
   }
   SetOwner();
}

//=====================================================================
//* destructor --------------------------------------------------------

U4ITTrack::~U4ITTrack()
{
}

//=====================================================================
//* MakeKalHits -------------------------------------------------------

void U4ITTrack::MakeKalHits(TObjArray &kalhits)
{
   TIter nextthit(this);
   U4ITTrackHit  *trackhitp = 0;

   while ((trackhitp = dynamic_cast<U4ITTrackHit *>(nextthit()))) {
      if (trackhitp->IsLocked()) continue;
      U4ITHit *hitp    = trackhitp->GetHitPtr();
      Int_t    layerno = hitp->GetLayerID();
      Double_t bfield  = dynamic_cast<U4ITKalDetector *>(GetTrackMakerPtr()
                       ->GetDetectorPtr())->GetBfield(layerno);
      const U4ITKalDetector &det = *static_cast<const U4ITKalDetector *>
                                   (GetTrackMakerPtr()->GetDetectorPtr());
      if (!det.IsFBOn() && hitp->GetSide()) continue;
      const U4VMeasLayer *msp = det.GetMeasLayerPtr(layerno);

      // =======================================================
      // Temporary treatment !!
      // -------------------------------------------------------
      const U4ITMeasLayer   *bmsp  = dynamic_cast<const U4ITMeasLayer *>(msp);
      const U4ITFBMeasLayer *fbmsp = dynamic_cast<const U4ITFBMeasLayer *>(msp);
      Double_t meas [2];
      Double_t dmeas[2];
      meas [0] = hitp->GetX1();        // cyllinder case : r_phi
      meas [1] = hitp->GetX2();        //                : z
      dmeas[0] = hitp->GetDX1();
      dmeas[1] = hitp->GetDX2();
      if (bmsp) {
         kalhits.Add(new U4ITKalHit(*trackhitp, *bmsp, meas, dmeas,  bfield));
      } else {
         kalhits.Add(new U4ITFBKalHit(*trackhitp, *fbmsp, meas, dmeas,  bfield));
      }
   // =================================================================
#ifdef __DEBUG__
      cerr.setf(ios::right);
      cerr.setf(ios::fixed);

      cerr << setw(4) << kalhits.GetEntries()
           << "(layr) = ("
           << setw(2) << layerno << ")"
           << " "
           << " (rphi, z) = (" 
           << setw(8) << setprecision(4) << meas [0] << " +/-"
           << setw(8) << setprecision(4) << dmeas[0] << ","
           << setw(8) << setprecision(4) << meas [1] << " +/-"
           << setw(8) << setprecision(4) << dmeas[1] << ")"
           << endl;
      cerr << setprecision(7);

      cerr.unsetf(ios::right);
      cerr.unsetf(ios::fixed);
#endif
   }
}



//=====================================================================
//* GetFirstSitePtr ---------------------------------------------------

TKalTrackSite *U4ITTrack::GetFirstSitePtr(Bool_t, TObjArray *)
{
   U4VTrack *stkp = GetSeedTrackPtr();
   if (!stkp) {
      // Self tracking mode that does not assume any seed track
      // Not implemented yet. Try following the TPC example
      // In that case, you should use dir and hitsp supplied as args
      cerr << " >>>>> U4ITTrack::GetFirstSitePtr >>>>>>>>> " << endl
           << "  Supply  a seed track, since self tracking " << endl
           << "  is not yet implemented                    " << endl;
      return 0;
   }
   TKalTrackSite *sitep = new TKalTrackSite(dynamic_cast<TKalTrackSite &>
                                     (stkp->GetKalTrack().GetCurSite()));
   sitep->SetHitOwner(kFALSE);
   return sitep;
}

//=====================================================================
//* DebugPrint --------------------------------------------------------

void U4ITTrack::DebugPrint(const Char_t *opt)
{
   cerr << "U4ITTrack::DebugPrint: NEntries = "  << GetEntries() << endl;
   TString opts = opt;
   if (opts.Contains("Detailed")) {
      if (!IsDone()) Update();
      TKalTrackState a = GetKalTrack().GetCurSite().GetCurState();

      cerr.setf(ios::right);
      cerr.setf(ios::fixed);
      cerr << " Type = " << setw(2) << GetTrackType()
           << " chi2 = " << setw(6) << setprecision(3) << GetChi2()
           << " ndf = " << setw(5) << GetNDF()
           << " t0 = " << setw(6) << setprecision(3) << GetT0()
           << " a = (" << setw(6) << setprecision(3) << a(0,0)
           << ","      << setw(6) << setprecision(3) << a(1,0)
           << ","      << setw(6) << setprecision(3) << a(2,0)
           << ","      << setw(6) << setprecision(3) << a(3,0)
           << ","      << setw(6) << setprecision(3) << a(4,0)
           << ")"
           << endl;
   }
}

//=====================================================================
//* GetNhits ----------------------------------------------------------

Int_t U4ITTrack::GetNhits()
{
   Int_t nhits = 0;
   TIter next(this);
   U4ITTrackHit *hitp = 0;
   while ((hitp = dynamic_cast<U4ITTrackHit *>(next()))) {
      if (!hitp->IsLocked()) {
         nhits++;
      }
   }
   return nhits;
} 

//=====================================================================
//* KalmanFilter ------------------------------------------------------

void U4ITTrack::KalmanFilter(Bool_t dir)
{
#ifdef __DEBUG__
   cerr << "U4ITTrack ::KalmanFilter is called..." << endl;
#endif
   SetTrackType(2);
   U4VTrack::KalmanFilter(dir);
}
