#ifndef __U4VTRACK__
#define __U4VTRACK__
//*************************************************************************
//* ==============
//* U4VTrack Class
//* ==============
//* (Description)
//*    Base class to all kinds of uranus charged tracks.
//*    Derived class should implement
//*      Copy C-tor if U4VTrack contains something other than U4VTrackHit
//*    and
//*      KalmanFilter (pure virtual)
//*    in any case.
//* (Requires)
//*    TObjArray
//* (Provides)
//*    U4VTrack    : Array containing "U4VTrackHit"s
//* (Update Record)
//*    2004/08/08  K.Fujii         Original version.
//*************************************************************************

#include "TObjArray.h"

#include "TAttLockable.h"
#include "TKalTrack.h"

class JSFHelicalTrack;
class U4VTrackMaker;

class U4VTrack : public TObjArray, public TAttLockable {
public:
  // ----------------------------------------------------------------------
  // C-tors and D-tor 
  // ----------------
  U4VTrack(U4VTrackMaker *trkm = 0);
  U4VTrack(const U4VTrack &tk); // must do deep copy
  virtual ~U4VTrack();

  // ----------------------------------------------------------------------
  // Sorting functions
  // -----------------
         virtual Bool_t	IsSortable ()                   const;
         virtual Int_t  Compare    (const TObject *obj) const;
         static  void   SetSortBy  (Char_t        *opt);
         static  void   SetSortBy  (Int_t          key);
  inline virtual Int_t  GetSortBy  () const { return fSortBy; }

  // ----------------------------------------------------------------------
  // Getters and setters
  // -------------------
  inline virtual Int_t     GetTrackNumber()   const   { return fTrackNo;  }
  inline virtual Int_t     GetTrackType  ();
  inline virtual Double_t  GetChi2       ();
  inline virtual Int_t     GetNDF        ();
  inline virtual Double_t  GetT0         ();
  inline virtual Int_t     GetNhits      ()   const;
  inline virtual Bool_t    IsDone        ()   const   { return fDone;     }

  inline virtual void      SetTrackNumber(Int_t    i) { fTrackNo   = i;   }
  inline virtual void      SetDone(Bool_t b =  kTRUE) { fDone      = b;   }
  inline virtual void      SetTrackType  (Int_t    i) { fTrackType = i;   }
  inline virtual void      SetT0         (Double_t d) { fT0        = d;   }

  // ----------------------------------------------------------------------
  // Service methods
  // ---------------
  inline virtual U4VTrackMaker   * GetTrackMakerPtr() const
                                         { return fTrackMakerPtr; }
  inline virtual JSFHelicalTrack * GetHelicalTrackPtr(Char_t *mode = "Kalman");
                                            // mode: (Helix, Kalman)
  inline virtual TKalTrack       & GetKalTrack();

  inline virtual U4VTrack  * GetSeedTrackPtr()  const       { return fSeedTrackPtr; }
  inline virtual void        SetSeedTrackPtr(U4VTrack *stp) { fSeedTrackPtr = stp;  }


         virtual void  Add(TObject *obj);

         virtual void  Update(const Char_t *mode = "Kalman"); // mode: (Helix, Kalman)
  inline virtual void  Undone()       { fDone = kFALSE; }

         virtual void  DebugPrint(const Char_t *opt = "Brief");
  inline virtual void  SetIPConstraint(Bool_t is = kTRUE)
                       { fIPConstraint = is; }

  inline virtual Bool_t IsIPConstraint() const { return fIPConstraint; }

  // ------------------------------------------------------------------
  // ===============
  //  Pure virtuals
  // ===============
  // ------------------------------------------------------------------
          virtual void MakeKalHits(TObjArray &kalhits) = 0;
          virtual TKalTrackSite *GetFirstSitePtr(Bool_t     dir,
                                                 TObjArray *kalhitsp = 0) = 0;
  // ------------------------------------------------------------------
protected:
         virtual void    SetNDF(Int_t    i);
         virtual void    SetChi2(Double_t d);

         virtual void    KalmanFilter(Bool_t dir = kIterBackward);
         virtual void    HelixFit();
         virtual void    SetJSFHelicalTrack(const TKalTrackSite  &site,
                                            const TKalTrackState &a,
                                            const TKalMatrix     &cov);

private:
  Int_t            fTrackNo;               //  track serial No.
  Bool_t 	   fDone;                  //  flag
  Int_t            fTrackType;             //  track type
  Double_t         fChi2;                  //  chi^2
  Int_t            fNDF;                   //  track type
  Double_t         fT0;                    //  T0
  Bool_t           fIPConstraint;          //  IP constraint flag
  TKalTrack        fKalTrack;              //! kalman track
  JSFHelicalTrack *fHelicalTrackPtr;       //  ptr to jsf helix
  U4VTrackMaker   *fTrackMakerPtr;         //  ptr to track maker
  U4VTrack        *fSeedTrackPtr;          //  ptr to seed track

  static Int_t     fSortBy;                //  (0,1,2) = by (Chi2,Time,SerialNo)
  
  ClassDef(U4VTrack, 1)		   // Basic Track Object
};

//-----------------------------------------------------------
//  inline function
//-----------------------------------------------------------

Int_t    U4VTrack::GetTrackType()
{
   if (!fDone) Update();
   return fTrackType;
}
Double_t U4VTrack::GetChi2()
{
   if (!fDone) Update();
   return fChi2;
}

Int_t U4VTrack::GetNDF()
{
   if (!fDone) Update();
   return fNDF;
}

Double_t U4VTrack::GetT0()
{
   if (!fDone) Update();
   return fT0;
}

Int_t U4VTrack::GetNhits() const
{
   return GetEntries();
}

TKalTrack & U4VTrack::GetKalTrack()
{
   if (!fDone) Update();
   return fKalTrack;
}

JSFHelicalTrack * U4VTrack::GetHelicalTrackPtr(Char_t *mode)
{
   if (!fDone) Update(mode);
   return fHelicalTrackPtr;
}

#endif
